%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% ALE source-level debugger, version 2.0
% Gerald Penn, July 1998
% using code from the SICStus source-level debugger, by Per Mildner
%
% For use only with SICStus Prolog 3.0.6 or higher
% To use without emacs, use noemacs/0, otherwise use XEmacs 19.16 or higher
%
% For installation instructions, see INSTALL
%
% TO RUN WITH XEMACS
% 0. Make sure that your EPROLOG environment variable is set to the command
%    that will run SICStus Prolog 3.0.6 or higher (not necessary if the command
%    is 'sicstus')
% 1. Run XEmacs
% 2. Run SICStus prolog as an inferior process with M-x run-prolog
% 3. ?- compile(debugger).  % ignore buffer_mode warning message
% 4. use dcompile_gram/1 to compile grammar files for debugging,
%    drec/1 for parsing, dgen/1 for generation, and dquery/1 for 
%    definite clause resolution
%
% TO RUN WITHOUT XEMACS
% 1. Run SICStus Prolog
% 2. ?- compile(debugger). % ignore buffer_mode warning message
% 3. ?- noemacs.           % you can put this at the end of debugger.pl too
% 4. use dcompile_gram/1 to compile grammar files for debugging,
%    drec/1 for parsing, and dquery/1 for definite clause resolution

:- dynamic ale_noemacs/0.

:- op(1100,fx,dquery).
:- op(1100,fx,dgen).
:- op(1100,fx,drec).

:- ['load_sicstus_debug'].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% emacs/0, noemacs/0:  enable/disable emacs-lisp control strings
emacs :-
  retractall(ale_noemacs),
  nl,write('emacs interface enabled'),
  nl.
noemacs :-
  assert(ale_noemacs),
  nl,write('emacs interface disabled'),
  nl.

emacs_format(Stream,Message,Vars) :-
  ale_noemacs -> true ; format(Stream,Message,Vars).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% dcompile_gram/1: compile file for debugging (currently must be all in one
%  file, except prolog hooks, which must be separate file(s))

dcompile_gram(File) :-
  retractall(ale_break(_)),
  retractall(ale_debugging(_)),
  retractall(ale_breakable_lines(_)),
  abolish_preds,
  abolish_lpreds,
  consult(File),
  pcompile_gram,
  label_items(File).

% pcompile_gram: compile type logic as before
pcompile_gram :-
  compile_sig,
  compile_fun,
  compile_cons,
  compile_logic,
  compile_dcs.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% drec/1:  Interactive parser
drec(Words) :-
  nl, write('STRING: '),
  nl, number_display(Words,0), 
  ttynl, drec(Words,Tag,SVs,Iqs),
  nl, write('CATEGORY: '),nl, ttyflush, 
  pp_fs(Tag-SVs,Iqs), 
  nl, query_proceed.

drec(Ws,Tag,SVs,IqsOut) :-
  clear,
  retractall(d_mode(_)),
  assert(d_mode(creep)),
  assert(parsing(Ws)),
  asserta(num(0)),
  reverse_count(Ws,[],WsRev,0,Length),
  d_build(WsRev,Length),
  edge(0,_,Length,Tag,SVs,IqsIn,_,_),
  extensionalise(Tag,SVs,IqsIn),
  check_inequal(IqsIn,IqsOut).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% dquery/1:  Interactive definite clause resolution
dquery(GoalDesc):-
  retractall(d_mode(_)),
  assert(d_mode(creep)),
  \+ \+
  (nl, label_query(GoalDesc,Functor,Arity,LDList),
   d_add_to_list_fresh(LDList,Args,[],IqsMid,0,1),
   on_exception(cut,d_solve_goal(Functor,Arity,Args,IqsMid,IqsMid2,0),fail),
   deref_list(Args,ArgsOut),
   extensionalise_list(ArgsOut,IqsMid2),
   check_inequal(IqsMid2,IqsOut),
   duplicates_list(ArgsOut,IqsOut,[],_,[],_,0,_),
   Goal =.. [Functor|ArgsOut],
   pp_goal(Goal,[],VisMid,0),
   nl, pp_iqs(IqsOut,VisMid,_,0),
   nl, query_proceed).

d_build([W|Ws],Right):-
  RightMinus1 is Right - 1,
  ( d_empty_cat(Right)
  ; d_lex(W,RightMinus1,Right)
  ; d_build(Ws,RightMinus1)
  ).
d_build([],_):-
  d_empty_cat(0).
d_build([],_) :-
  quiet_interpreter,  % if quiet mode was on, turn it off.
  !,secret_verbose.
d_build([],_).

d_empty_cat(Node) :-
  current_predicate(lempty,lempty(_,_)),
  lempty(OpLine,LDesc),
  d_call_fail(0,_,empty,OpLine),
  on_exception(ale_fail,(d_add_to(LDesc,Tag,bot,[],IqsIn,1,empty),
                         d_exit_redo(0,empty,OpLine)),
               fail),
  fully_deref_prune(Tag,bot,TagOut,SVsOut,IqsIn,IqsOut),
  d_add_edge(Node,Node,TagOut,SVsOut,IqsOut,[],empty,0).

d_add_to(lvar(X,XName,XLine),Ref2,SVs2,IqsIn,IqsOut,Skip,Loc):-
  !,    % also instantiates variable first arguments
  d_call_fail(Skip,_,unify(Loc,XName,Ref2-SVs2,IqsIn,X),XLine),
  on_exception(ale_fail,(X = Ref1-bot,
                         deref(Ref1,bot,Ref3,SVs3),
                         u(SVs2,SVs3,Ref2,Ref3,IqsIn,IqsOut),
                         d_exit_redo(Skip,unify(Loc,XName,Ref2-SVs2,IqsOut,X),
                                     XLine)),
               fail).
d_add_to(lelist(LBrackLine),Ref,SVs,IqsIn,IqsOut,Skip,Loc):-
  !, d_add_to(ltype(e_list,LBrackLine),Ref,SVs,IqsIn,IqsOut,Skip,Loc).
d_add_to(lnelist(HeadLine,TailLine,LH,LT),Ref,SVs,IqsIn,IqsOut,Skip,Loc):-
  !, d_add_to((lfeat(hd,HeadLine):LH,lfeat(tl,TailLine):LT),Ref,SVs,
              IqsIn,IqsOut,Skip,Loc).
d_add_to(leq(Path1,Path2,OpLine,LP1,LP2),Tag,SVs,IqsIn,IqsOut,Skip,Loc):-
  !, d_call_fail(Skip,NewSkip,eq(Loc,Path1,Path2,Tag-SVs,IqsIn),OpLine),
  on_exception(ale_fail,
    (d_pathval(LP1,Tag,SVs,TagAtPath1,SVsAtPath1,IqsIn,IqsMid,NewSkip),
     deref(Tag,SVs,TagMid,SVsMid),
     d_pathval(LP2,TagMid,SVsMid,TagAtPath2,SVsAtPath2,IqsMid,IqsMid2,NewSkip),
     u(SVsAtPath1,SVsAtPath2,TagAtPath1,TagAtPath2,IqsMid2,IqsOut),
     d_exit_redo(Skip,eq(Loc,Path1,Path2,Tag-SVs,IqsOut),OpLine)),
    fail).
d_add_to(lineq(OpLine,LabelledDesc),Tag,SVs,IqsIn,IqsOut,Skip,Loc):-
  !, d_call_fail(Skip,NewSkip,ineq_add(Loc,Tag-SVs,IqsIn),OpLine),
  on_exception(ale_fail,(d_add_to(LabelledDesc,Tag2,bot,IqsIn,IqsMid,NewSkip,
                                  ineq),
                         check_inequal_conjunct(
                                    ineq(Tag,SVs,Tag2,bot,done),IqOut,Result),
                         ( (Result = succeed) ->
                             IqsOut = IqsMid
                         ; IqOut \== done,
                           IqsOut = [IqOut|IqsMid]),
                         d_exit_redo(Skip,ineq_add(Loc,Tag-SVs,IqsOut),
                                     OpLine)),
               fail).
d_add_to(lfeat(Feat,FLine):LabelledDesc,Ref,SVs,IqsIn,IqsOut,Skip,Loc):-
  !, d_call_fail(Skip,NewSkip,featval(Loc,Feat,Ref-SVs,IqsIn),FLine),
  on_exception(ale_fail,
          (featval(Feat,SVs,Ref,FSatFeat,IqsIn,IqsMid), 
           deref(FSatFeat,RefatFeat,SVsatFeat),
           d_add_to(LabelledDesc,RefatFeat,SVsatFeat,IqsMid,IqsOut,NewSkip,
                    feat(Feat)),
           d_exit_redo(Skip,featval(Loc,Feat,Ref-SVs,IqsOut),FLine)),
         fail).
d_add_to((LabelledDesc1,LabelledDesc2),Ref,SVs,IqsIn,IqsOut,Skip,Loc):-
  !, d_add_to(LabelledDesc1,Ref,SVs,IqsIn,IqsMid,Skip,Loc),
  deref(Ref,SVs,Ref2,SVs2), 
  d_add_to(LabelledDesc2,Ref2,SVs2,IqsMid,IqsOut,Skip,Loc).
d_add_to((LabelledDesc1;LabelledDesc2),Ref,SVs,IqsIn,IqsOut,Skip,Loc):-
  !, 
  ( d_add_to(LabelledDesc1,Ref,SVs,IqsIn,IqsOut,Skip,Loc)
  ; d_add_to(LabelledDesc2,Ref,SVs,IqsIn,IqsOut,Skip,Loc)
  ).
d_add_to(lmac(MacroName,OpLine),Ref,SVs,IqsIn,IqsOut,Skip,Loc):-
  !, d_call_fail(Skip,NewSkip,macro(MacroName,Ref-SVs,IqsIn),OpLine),
  on_exception(ale_fail,
                   (lmacro(MacroName,LabelledMacroDesc),
                    d_add_to(LabelledMacroDesc,Ref,SVs,IqsIn,IqsOut,NewSkip,
                             Loc),
                    d_exit_redo(Skip,macro(MacroName,Ref-SVs,IqsOut),OpLine)),
               fail).
d_add_to(ltype(Type,TLine),Ref,SVs,IqsIn,IqsOut,Skip,Loc):-
  type(Type),
  !, d_call_fail(Skip,_,type(Loc,Type,Ref-SVs,IqsIn),TLine),
  on_exception(ale_fail,(add_to_type(Type,SVs,Ref,IqsIn,IqsOut),
                        d_exit_redo(Skip,type(Loc,Type,Ref-SVs,IqsOut),TLine)),
               fail).
d_add_to(latom(Atom,ALine),Ref,SVs,IqsIn,IqsOut,Skip,Loc):-
  !, d_call_fail(Skip,_,atom(Loc,Atom,Ref-SVs,IqsIn),ALine),
  on_exception(ale_fail,(add_to_type((a_ Atom),SVs,Ref,IqsIn,IqsOut),
                        d_exit_redo(Skip,atom(Loc,Atom,Ref-SVs,IqsOut),ALine)),
               fail).
d_add_to(lfun(Fun,lcomp(Functor,Arity,FunLine,LDList)),Ref,SVs,IqsIn,IqsOut,
         Skip,Loc) :-
  fun(Fun),
  !, d_call_fail(Skip,NewSkip,fun(Functor,Arity,Ref-SVs,IqsIn),FunLine),
  on_exception(ale_fail,
                (d_fsolve(Functor,Arity,LDList,Ref,SVs,IqsIn,IqsOut,NewSkip,
                          Loc),
                 d_exit_redo(Skip,fun(Functor,Arity,Ref-SVs,IqsOut),FunLine)),
               fail).

d_pathval([],Tag,SVs,Tag,SVs,Iqs,Iqs,_).
d_pathval([lfeat(Feat,FLine)|LPath],Tag,SVs,TagOut,SVsOut,IqsIn,IqsOut,Skip):-
  d_call_fail(Skip,NewSkip,featval(path,Feat,Tag-SVs,IqsIn),FLine),
  on_exception(ale_fail,
         (featval(Feat,SVs,Tag,FSMid,IqsIn,IqsMid),
          deref(FSMid,TagMid,SVsMid),
          d_pathval(LPath,TagMid,SVsMid,TagOut,SVsOut,IqsMid,IqsOut,NewSkip),
          d_exit_redo(Skip,featval(path,Feat,Tag-SVs,IqsOut),FLine)),
         fail).

d_lex(Word,Left,Right) :-
  'l--->'(WordStart,WordLine,LabelledDesc),
  d_call_fail(0,Skip,lex(Word,WordStart),WordLine),
  on_exception(ale_fail,
     (d_add_to(LabelledDesc,TagStart,bot,[],IqsMid,Skip,lex),
      curr_lex_rule_depth(Max),
      d_lex_close(0,Max,WordStart,TagStart,bot,Word,TagMid,SVsMid,IqsMid,
                  IqsMid2,Skip),
      d_exit_redo(0,lex(Word,WordStart),WordLine)),
     fail),
  fully_deref_prune(TagMid,SVsMid,Tag,SVs,IqsMid2,IqsOut),
  d_add_edge(Left,Right,Tag,SVs,IqsOut,[],lexicon,0).

d_lex_close(_,_,Word,Tag,SVs,Word,Tag,SVs,Iqs,Iqs,_).
d_lex_close(N,Max,WordIn,TagIn,SVsIn,WordOut,TagOut,SVsOut,IqsIn,IqsOut,Skip):-
  current_predicate(llr,llr(_,_,_)),
  N < Max,
  d_lex_rule(WordIn,TagIn,SVsIn,WordMid,TagMid,SVsMid,IqsIn,IqsMid,Skip),
  NPlus1 is N + 1,
  d_lex_close(NPlus1,Max,WordMid,TagMid,SVsMid,WordOut,TagOut,SVsOut,IqsMid,
              IqsOut,Skip).

d_lex_rule(WordIn,TagIn,SVsIn,WordOut,TagOut,SVsOut,IqsIn,IqsOut,Skip) :-
  llr(RuleName,InLine,llrbody(LNMBody,LMorphs)),
  d_call_fail(Skip,NewSkip,lr(RuleName,WordIn,TagIn-SVsIn,IqsIn),InLine),
  on_exception(ale_fail,(( LNMBody = llrmap(LDIn,LDOut) ->
                            LabelledCond = ltrue
                         ;  LNMBody = lif(llrmap(LDIn,LDOut),LabelledCond)),
                         deref(TagIn,SVsIn,DTagIn,DSVsIn),
                         d_add_to(LDIn,DTagIn,DSVsIn,IqsIn,IqsMid,NewSkip,lrin),
                         d_solve(LabelledCond,IqsMid,IqsMid2,NewSkip),
                         d_add_to(LDOut,TagMid,bot,IqsMid2,IqsMid3,NewSkip,
                                  lrout),
                         d_morphs(LMorphs,WordIn,WordOut,NewSkip),
                         d_exit_redo(Skip,lr(RuleName,WordIn,TagMid-bot,
                                             IqsMid3),
                                     InLine)),
               fail),
  fully_deref_prune(TagMid,bot,TagOut,SVsOut,IqsMid3,IqsOut).

d_morphs(lmorph_seq(LMorph,LMorphs),WordIn,WordOut,Skip) :-
  !,(d_morph(LMorph,WordIn,WordOut,Skip) ->
       true
    ; d_morphs(LMorphs,WordIn,WordOut,Skip)).
d_morphs(LMorph,WordIn,WordOut,Skip) :-
  d_morph(LMorph,WordIn,WordOut,Skip).

d_morph(lmorph_cond(WhenLine,lmorph(OpLine,StrIn,StrOut),Cond),WordIn,WordOut,
        Skip) :-
  !, d_call_fail(Skip,NewSkip,morph_input(WordIn),OpLine),
  on_exception(ale_fail,(name(WordIn,CodesIn),
                         make_char_list(CodesIn,CharsIn),
                         morph_pattern(StrIn,CharsIn),
                         d_call_fail(NewSkip,_,when,WhenLine),
                         on_exception(ale_fail,((d_mode(creep) ->
                                                   trace
                                                ; true),
                                                call(Cond),
                                                notrace,
                                           d_exit_redo(NewSkip,when,WhenLine)),
                                      fail),
                         morph_pattern(StrOut,CharsOut),
                         make_char_list(CodesOut,CharsOut),
                         name(WordOut,CodesOut),
                        d_exit_redo(Skip,morph_output(WordIn,WordOut),OpLine)),
               fail).
d_morph(lmorph(OpLine,StrIn,StrOut),WordIn,WordOut,Skip) :-
  d_call_fail(Skip,_,morph_input(WordIn),OpLine),
  on_exception(ale_fail,(name(WordIn,CodesIn),
                         make_char_list(CodesIn,CharsIn),
                         morph_pattern(StrIn,CharsIn),
                         morph_pattern(StrOut,CharsOut),
                         make_char_list(CodesOut,CharsOut),
                         name(WordOut,CodesOut),
                        d_exit_redo(Skip,morph_output(WordIn,WordOut),OpLine)),
               fail).

d_solve(((LGoal1,LGoal2),LGoal3),IqsIn,IqsOut,Skip) :-
  !,d_solve((LGoal1,(LGoal2,LGoal3)),IqsIn,IqsOut,Skip).
d_solve(((LGoal1;LGoal2),LGoal3),IqsIn,IqsOut,Skip) :-
  !,d_solve(((LGoal1,LGoal3);(LGoal2,LGoal3)),IqsIn,IqsOut,Skip).
d_solve((lneg(NLine,LGoal1),LGoal2),IqsIn,IqsOut,Skip) :-
  !,d_call_fail(Skip,NewSkip,neg,NLine),
  on_exception(ale_fail,(\+ d_solve(LGoal1,IqsIn,_,NewSkip),
                         d_exit_redo(Skip,neg,NLine)),
               fail),
  d_solve(LGoal2,IqsIn,IqsOut,Skip).
d_solve((lexteq(OpLine,LD1,LD2),LGoal),IqsIn,IqsOut,Skip) :-
  !,d_call_fail(Skip,NewSkip,exteq,OpLine),
  on_exception(ale_fail,(d_add_to(LD1,Tag1,bot,IqsIn,IqsMid,NewSkip,left),
                         d_add_to(LD2,Tag2,bot,IqsMid,IqsMid2,NewSkip,right),
                         deref(Tag1,bot,DTag1,DSVs1),
                         deref(Tag2,bot,DTag2,DSVs2),
                        extensionalise_list([DTag1-DSVs1,DTag2-DSVs2],IqsMid2),
                         check_inequal(IqsMid2,IqsMid3),
                         deref(DTag1,DSVs1,Tag1Out,_),
                         deref(DTag2,DSVs2,Tag2Out,_),
                         Tag1Out == Tag2Out,
                         d_exit_redo(Skip,exteq,OpLine)),
               fail),
  d_solve(LGoal,IqsMid3,IqsOut,Skip).
d_solve((ltrue,LGoal),IqsIn,IqsOut,Skip) :-
  !,d_solve(LGoal,IqsIn,IqsOut,Skip).
d_solve((lcut,LGoal),IqsIn,IqsOut,Skip) :-
  !,
  ( d_solve(LGoal,IqsIn,IqsOut,Skip)
  ; raise_exception(cut)).
d_solve((lprolog(OpLine,PGoal),LGoal),IqsIn,IqsOut,Skip) :-
  !,d_call_fail(Skip,_,prolog(PGoal),OpLine),
  on_exception(ale_fail,((d_mode(creep) ->
                           trace
                         ; true),
                         call(PGoal),
                         notrace,
                         d_exit_redo(Skip,prolog(PGoal),OpLine)),
               fail),
  d_solve(LGoal,IqsIn,IqsOut,Skip).
d_solve((AtLGoal,LGoal),IqsIn,IqsOut,Skip) :-
  !,d_solve(AtLGoal,IqsIn,IqsMid,Skip),
  d_solve(LGoal,IqsMid,IqsOut,Skip).
d_solve((LGoal1;LGoal2),IqsIn,IqsOut,Skip) :-
  !,( d_solve(LGoal1,IqsIn,IqsOut,Skip)
    ; d_solve(LGoal2,IqsIn,IqsOut,Skip)).
d_solve(lneg(NLine,LGoal),Iqs,Iqs,Skip) :-
  !,d_call_fail(Skip,NewSkip,neg,NLine),
  on_exception(ale_fail,(\+ d_solve(LGoal,Iqs,_,NewSkip),
                         d_exit_redo(Skip,neg,NLine)),
               fail).
d_solve(lexteq(OpLine,LD1,LD2),IqsIn,IqsOut,Skip) :-
  !,d_call_fail(Skip,NewSkip,exteq,OpLine),
  on_exception(ale_fail,(d_add_to(LD1,Tag1,bot,IqsIn,IqsMid,NewSkip,left),
                         d_add_to(LD2,Tag2,bot,IqsMid,IqsMid2,NewSkip,right),
                         deref(Tag1,bot,DTag1,DSVs1),
                         deref(Tag2,bot,DTag2,DSVs2),
                        extensionalise_list([DTag1-DSVs1,DTag2-DSVs2],IqsMid2),
                         check_inequal(IqsMid2,IqsOut),
                         deref(DTag1,DSVs1,Tag1Out,_),
                         deref(DTag2,DSVs2,Tag2Out,_),
                         Tag1Out == Tag2Out,
                         d_exit_redo(Skip,exteq,OpLine)),
               fail).
d_solve(ltrue,Iqs,Iqs,_) :-
  !.
d_solve(lcut,Iqs,Iqs,_) :-
  !,
  ( true
  ; raise_exception(cut)).
d_solve(lprolog(OpLine,PGoal),Iqs,Iqs,Skip) :-
  !,d_call_fail(Skip,_,prolog(PGoal),OpLine),
  on_exception(ale_fail,((d_mode(creep) ->
                           trace
                         ; true),
                         call(PGoal),
                         notrace,
                         d_exit_redo(Skip,prolog(PGoal),OpLine)),
               fail).
d_solve(lcomp(Functor,Arity,FunLine,LabelledDescList),IqsIn,IqsOut,Skip) :-
  d_call_fail(Skip,NewSkip,comp(Functor,Arity),FunLine),
  on_exception(ale_fail,
      (d_add_to_list_fresh(LabelledDescList,Args,IqsIn,IqsMid,NewSkip,1),
       on_exception(cut,d_solve_goal(Functor,Arity,Args,IqsMid,IqsOut,NewSkip),
                    fail),
       d_exit_redo(Skip,comp(Functor,Arity),FunLine)),
      fail).

d_solve_goal(Rel,Arity,Args,IqsIn,IqsOut,Skip) :-
  current_predicate(lif,lif(_,_)),
  lif(lcomp(Rel,Arity,FunLine,LabelledDescs),LabelledBody),
  d_call_fail(Skip,NewSkip,comp_clause(Rel,Arity),FunLine),
  on_exception(ale_fail,
               (d_add_to_list(LabelledDescs,Args,IqsIn,IqsMid,NewSkip,1),
                check_inequal(IqsMid,IqsMid2),
                d_solve(LabelledBody,IqsMid2,IqsOut,NewSkip),
                d_exit_redo(Skip,comp_clause(Rel,Arity),FunLine)),
               fail).

d_fsolve(Rel,Arity,LArgDescs1,Ref,SVs,IqsIn,IqsOut,Skip,Loc) :-
  current_predicate('l+++>','l+++>'(_,_)),
  d_add_to_list_fresh(LArgDescs1,Args,IqsIn,IqsMid,Skip,1),
  'l+++>'(lcomp(Rel,Arity,FunLine,LArgDescs2),LabelledResult),
  d_call_fail(Skip,NewSkip,fun_clause(Rel,Arity),FunLine),
  on_exception(ale_fail,
               (d_add_to_list(LArgDescs2,Args,IqsMid,IqsMid2,NewSkip,1),
                check_inequal(IqsMid2,IqsMid3),
                d_add_to(LabelledResult,Ref,SVs,IqsMid3,IqsOut,NewSkip,Loc),
                d_exit_redo(Skip,fun_clause(Rel,Arity),FunLine)),
               fail).

d_add_to_list_fresh([],[],Iqs,Iqs,_,_).
d_add_to_list_fresh([LDesc|LDescs],[Tag-bot|FSs],IqsIn,IqsOut,Skip,N) :-
  d_add_to(LDesc,Tag,bot,IqsIn,IqsMid,Skip,arg(N)),
  NewN is N + 1,
  d_add_to_list_fresh(LDescs,FSs,IqsMid,IqsOut,Skip,NewN).

d_add_to_list([],[],Iqs,Iqs,_,_).
d_add_to_list([LDesc|LDescs],[FS|FSs],IqsIn,IqsOut,Skip,N) :-
  deref(FS,Tag,SVs),
  d_add_to(LDesc,Tag,SVs,IqsIn,IqsMid,Skip,arg(N)),
  NewN is N + 1,
  d_add_to_list(LDescs,FSs,IqsMid,IqsOut,Skip,NewN).

d_add_edge(Left,Right,TagOut,SVsOut,IqsOut,Dtrs,RuleName,Skip):-
  (subsumed(Left,Right,TagOut,SVsOut,IqsOut,Dtrs,RuleName),
   !,fail
  ;edge_assert(Left,Right,TagOut,SVsOut,IqsOut,Dtrs,RuleName,N),
   N >= 0,  % edge_assert asserted the edge
   d_edge_announce_add(N,Left,Right,RuleName),
   !,d_call_fail(Skip,NewSkip,rule_close,none),
   on_exception(ale_fail,(d_rule(TagOut,SVsOut,IqsOut,Left,Right,N,NewSkip),
                          d_exit_redo(Skip,rule_close,none)),
                fail)).

d_edge_announce_add(N,Left,Right,RuleName) :-
  format('Edge added: Number: ~w, Left: ~w, Right: ~w, Rule: ~w~n',
         [N,Left,Right,RuleName]),
  ttyflush.

d_edge_announce_retrieve(N,Left,Right,RuleName) :-
  ((d_mode(creep)
   ;d_mode(line(_))) ->
    format('Edge retrieved: Number: ~w, Left: ~w, Right: ~w, Rule: ~w~n',
           [N,Left,Right,RuleName]),
    ttyflush
  ; true).

d_rule(Tag,SVs,Iqs,Left,Right,N,Skip) :-
  current_predicate(lrule,lrule(_,_,_,_)),
  lrule(RuleName,RuleLine,LabelledMother,LabelledRuleBody),
  d_call_fail(Skip,NewSkip,rule(RuleName),RuleLine),
  on_exception(ale_fail,
        (d_add_dtrs(LabelledRuleBody,Tag,SVs,Iqs,Left,Right,N,LabelledMother,
                    RuleName,NewSkip),
         d_exit_redo(Skip,rule(RuleName),RuleLine)),
        fail).

d_add_dtrs((lcat(LDtr),Rest),Tag,SVs,IqsIn,Left,Right,N,LMother,RuleName,
           Skip) :-
  !,d_add_to(LDtr,Tag,SVs,IqsIn,IqsMid,Skip,edge),
  check_inequal(IqsMid,IqsOut),
  d_add_dtrs_rest(Rest,Left,Right,IqsOut,LMother,[N|DtrsRest],DtrsRest,
                  RuleName,Skip).
d_add_dtrs((lsem_head(LDtr),Rest),Tag,SVs,IqsIn,Left,Right,N,LMother,RuleName,
           Skip) :-
  !,d_add_to(LDtr,Tag,SVs,IqsIn,IqsMid,Skip,edge),
  check_inequal(IqsMid,IqsOut),
  d_add_dtrs_rest(Rest,Left,Right,IqsOut,LMother,[N|DtrsRest],DtrsRest,
                  RuleName,Skip).
d_add_dtrs((lcats(LDtrs),Rest),Tag,SVs,IqsIn,Left,Right,N,LMother,RuleName,
           Skip) :-
  !,d_add_to(LDtrs,Tag2,bot,IqsIn,IqsMid,Skip,dtrlist),
  check_inequal(IqsMid,IqsMid2),
  deref(Tag2,bot,_,DescSVs),
  DescSVs =.. [Sort|Vs], 
  ((Sort == e_list) ->
     d_add_dtrs(Rest,Tag,SVs,IqsMid2,Left,Right,N,LMother,RuleName,Skip)
  ; (d_match_list(Sort,Vs,Tag,SVs,Right,N,Dtrs,DtrsRest,NextRight,
                  IqsMid2,IqsOut),
     d_add_dtrs_rest(Rest,Left,NextRight,IqsOut,LMother,Dtrs,DtrsRest,
                     RuleName,Skip))).
d_add_dtrs((lgoal(LGoal),Rest),Tag,SVs,IqsIn,Left,Right,N,LMother,RuleName,
           Skip):-
  !, d_solve(LGoal,IqsIn,IqsMid,Skip),
  d_add_dtrs(Rest,Tag,SVs,IqsMid,Left,Right,N,LMother,RuleName,Skip).
d_add_dtrs((lsem_goal(LGoal),Rest),Tag,SVs,IqsIn,Left,Right,N,LMother,RuleName,
           Skip):-
  !, d_solve(LGoal,IqsIn,IqsMid,Skip),
  d_add_dtrs(Rest,Tag,SVs,IqsMid,Left,Right,N,LMother,RuleName,Skip).
d_add_dtrs(lcat(LDtr),Tag,SVs,IqsIn,Left,Right,N,LMother,RuleName,Skip):-
  !,d_add_to(LDtr,Tag,SVs,IqsIn,IqsMid,Skip,edge),
  check_inequal(IqsMid,IqsMid2),
  d_add_to(LMother,Tag2,bot,IqsMid2,IqsMid3,Skip,mother),
  fully_deref_prune(Tag2,bot,Tag2Out,SVs2Out,IqsMid3,IqsOut),
  d_add_edge(Left,Right,Tag2Out,SVs2Out,IqsOut,[N],RuleName,Skip).
d_add_dtrs(lsem_head(LDtr),Tag,SVs,IqsIn,Left,Right,N,LMother,RuleName,Skip):-
  !,d_add_to(LDtr,Tag,SVs,IqsIn,IqsMid,Skip,edge),
  check_inequal(IqsMid,IqsMid2),
  d_add_to(LMother,Tag2,bot,IqsMid2,IqsMid3,Skip,mother),
  fully_deref_prune(Tag2,bot,Tag2Out,SVs2Out,IqsMid3,IqsOut),
  d_add_edge(Left,Right,Tag2Out,SVs2Out,IqsOut,[N],RuleName,Skip).
d_add_dtrs(lcats(LDtrs),Tag,SVs,IqsIn,Left,Right,N,LMother,RuleName,Skip) :-
  !,d_add_to(LDtrs,Tag3,bot,IqsIn,IqsMid,Skip,dtrlist),
  check_inequal(IqsMid,IqsMid2),
  deref(Tag3,bot,_,DescSVs),
  DescSVs =.. [Sort|Vs],
  ((Sort == e_list) ->
     fail
  ; (d_match_list(Sort,Vs,Tag,SVs,Right,N,Dtrs,[],NextRight,
                  IqsMid2,IqsMid3),
     d_add_to(LMother,Tag2,bot,IqsMid3,IqsOut,Skip,mother),
     fully_deref_prune(Tag2,bot,Tag2Out,SVs2Out,IqsMid3,IqsOut),
     d_add_edge(Left,NextRight,Tag2Out,SVs2Out,IqsOut,Dtrs,RuleName,Skip))).

d_add_dtrs_rest((lcat(LDtr),Rest),Left,Right,IqsMid,LMother,
                PrevDtrs,[N|DtrsRest],RuleName,Skip):-
  !,edge(Right,N,NewRight,Tag,SVs,EdgeIqs,_,ERN),
  d_edge_announce_retrieve(N,Right,NewRight,ERN),
  d_add_to(LDtr,Tag,SVs,IqsMid,IqsMid2,Skip,edge),
  append(IqsMid2,EdgeIqs,IqsMid3),
  check_inequal(IqsMid3,IqsOut),
  d_add_dtrs_rest(Rest,Left,NewRight,IqsOut,LMother,PrevDtrs,DtrsRest,
                  RuleName,Skip).
d_add_dtrs_rest((lsem_head(LDtr),Rest),Left,Right,IqsMid,LMother,
                PrevDtrs,[N|DtrsRest],RuleName,Skip):-
  !,edge(Right,N,NewRight,Tag,SVs,EdgeIqs,_,ERN),
  d_edge_announce_retrieve(N,Right,NewRight,ERN),
  d_add_to(LDtr,Tag,SVs,IqsMid,IqsMid2,Skip,edge),
  append(IqsMid2,EdgeIqs,IqsMid3),
  check_inequal(IqsMid3,IqsOut),
  d_add_dtrs_rest(Rest,Left,NewRight,IqsOut,LMother,PrevDtrs,DtrsRest,
                  RuleName,Skip).
d_add_dtrs_rest((lcats(LDtrs),Rest),Left,Right,IqsMid,LMother,
                PrevDtrs,DtrsRest,RuleName,Skip) :-
  !,d_add_to(LDtrs,Tag,bot,IqsMid,IqsMid2,Skip,dtrlist),
  check_inequal(IqsMid2,IqsMid3),
  deref(Tag,bot,_,SVs),
  SVs =.. [Sort|Vs],
  d_match_list_rest(Sort,Vs,Right,NewRight,DtrsRest,DtrsRest2,IqsMid3,IqsOut),
  d_add_dtrs_rest(Rest,Left,NewRight,IqsOut,LMother,PrevDtrs,DtrsRest2,
                  RuleName,Skip).
d_add_dtrs_rest((lgoal(LGoal),Rest),Left,Right,IqsMid,LMother,
                PrevDtrs,DtrsRest,RuleName,Skip):-
  !, d_solve(LGoal,IqsMid,IqsMid2,Skip),
  d_add_dtrs_rest(Rest,Left,Right,IqsMid2,LMother,PrevDtrs,DtrsRest,RuleName,
                  Skip).
d_add_dtrs_rest((lsem_goal(LGoal),Rest),Left,Right,IqsMid,LMother,
                PrevDtrs,DtrsRest,RuleName,Skip):-
  !, d_solve(LGoal,IqsMid,IqsMid2,Skip),
  d_add_dtrs_rest(Rest,Left,Right,IqsMid2,LMother,PrevDtrs,DtrsRest,RuleName,
                  Skip).
d_add_dtrs_rest(lcat(LDtr),Left,Right,IqsMid,LMother,PrevDtrs,[N],RuleName,
                Skip):-
  !,edge(Right,N,NewRight,Tag,SVs,EdgeIqs,_,ERN),
  d_edge_announce_retrieve(N,Right,NewRight,ERN),
  d_add_to(LDtr,Tag,SVs,IqsMid,IqsMid2,Skip,edge),
  append(IqsMid2,EdgeIqs,IqsMid3),
  check_inequal(IqsMid3,IqsMid4),
  d_add_to(LMother,Tag2,bot,IqsMid4,IqsMid5,Skip,mother),
  fully_deref_prune(Tag2,bot,Tag2Out,SVs2Out,IqsMid5,IqsOut),
  d_add_edge(Left,NewRight,Tag2Out,SVs2Out,IqsOut,PrevDtrs,RuleName,Skip).
d_add_dtrs_rest(lsem_head(LDtr),Left,Right,IqsMid,LMother,PrevDtrs,[N],
                RuleName,Skip):-
  !,edge(Right,N,NewRight,Tag,SVs,EdgeIqs,_,ERN),
  d_edge_announce_retrieve(N,Right,NewRight,ERN),
  d_add_to(LDtr,Tag,SVs,IqsMid,IqsMid2,Skip,edge),
  append(IqsMid2,EdgeIqs,IqsMid3),
  check_inequal(IqsMid3,IqsMid4),
  d_add_to(LMother,Tag2,bot,IqsMid4,IqsMid5,Skip,mother),
  fully_deref_prune(Tag2,bot,Tag2Out,SVs2Out,IqsMid5,IqsOut),
  d_add_edge(Left,NewRight,Tag2Out,SVs2Out,IqsOut,PrevDtrs,RuleName,Skip).
d_add_dtrs_rest(lcats(LDtrs),Left,Right,IqsMid,LMother,PrevDtrs,DtrsRest,
                RuleName,Skip) :-
  !,d_add_to(LDtrs,Tag,bot,IqsMid,IqsMid2,Skip,dtrlist),
  check_inequal(IqsMid2,IqsMid3),
  deref(Tag,bot,_,SVs),
  SVs =.. [Sort|Vs],
  d_match_list_rest(Sort,Vs,Right,NewRight,DtrsRest,[],IqsMid3,IqsMid4),
  d_add_to(LMother,Tag2,bot,IqsMid4,IqsMid5,Skip,mother),
  fully_deref_prune(Tag2,bot,Tag2Out,SVs2Out,IqsMid5,IqsOut),
  d_add_edge(Left,NewRight,Tag2Out,SVs2Out,IqsOut,PrevDtrs,RuleName,Skip).
d_add_dtrs_rest(lgoal(LGoal),Left,Right,IqsMid,LMother,PrevDtrs,[],RuleName,
                Skip):-
  !, d_solve(LGoal,IqsMid,IqsMid2,Skip),
  d_add_to(LMother,Tag2,bot,IqsMid2,IqsMid3,Skip,mother),
  fully_deref_prune(Tag2,bot,Tag2Out,SVs2Out,IqsMid3,IqsOut),
  d_add_edge(Left,Right,Tag2Out,SVs2Out,IqsOut,PrevDtrs,RuleName,Skip).
d_add_dtrs_rest(lsem_goal(LGoal),Left,Right,IqsMid,LMother,PrevDtrs,[],
                RuleName,Skip):-
  !, d_solve(LGoal,IqsMid,IqsMid2,Skip),
  d_add_to(LMother,Tag2,bot,IqsMid2,IqsMid3,Skip,mother),
  fully_deref_prune(Tag2,bot,Tag2Out,SVs2Out,IqsMid3,IqsOut),
  d_add_edge(Left,Right,Tag2Out,SVs2Out,IqsOut,PrevDtrs,RuleName,Skip).

d_match_list(Sort,[HdFS,TlFS],Tag,SVs,Right,N,[N|DtrsMid],DtrsRest,NextRight,
             IqsIn,IqsOut) :-
  sub_type(ne_list,Sort),
  !,ud(HdFS,Tag,SVs,IqsIn,IqsMid),
  check_inequal(IqsMid,IqsMid2),
  deref(TlFS,_,TlSVs),
  TlSVs =.. [TlSort|TlVs],  % a_ correctly causes error in recursive call
  d_match_list_rest(TlSort,TlVs,Right,NextRight,DtrsMid,DtrsRest,
                    IqsMid2,IqsOut).
d_match_list(Sort,_,_,_,_,_,_,_,_,_,_) :-
  error_msg((nl,write('error: cats> value with sort, '),write(Sort),
            write(' is not a valid list argument'))).

d_match_list_rest(e_list,_,Right,Right,DtrsRest,DtrsRest,Iqs,Iqs) :- 
  !.
d_match_list_rest(Sort,[HdFS,TlFS],Right,NewRight,[N|DtrsRest],DtrsRest2,IqsIn,
                IqsOut) :-
  sub_type(ne_list,Sort),
  !,edge(Right,N,MidRight,Tag,SVs,EdgeIqs,_,ERN),
  d_edge_announce_retrieve(N,Right,MidRight,ERN),
  ud(HdFS,Tag,SVs,IqsIn,IqsMid),
  append(IqsMid,EdgeIqs,IqsMid2),
  check_inequal(IqsMid2,IqsMid3),
  deref(TlFS,_,TlSVs),
  TlSVs =.. [TlSort|TlVs],  % a_ correctly causes error in recursive call
  d_match_list_rest(TlSort,TlVs,MidRight,NewRight,DtrsRest,DtrsRest2,
                    IqsMid3,IqsOut).
d_match_list_rest(Sort,_,_,_,_,_,_,_) :-
  error_msg((nl,write('error: cats> value with sort, '),write(Sort),
            write(' is not a valid list argument'))).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% dgen/1,3: Interactive generator
dgen(Desc) :-
 retractall(d_mode(_)),
 assert(d_mode(creep)),
 label_query_arg(Desc,LDesc),
 d_add_to(LDesc,Tag,bot,[],IqsIn,0,query_desc),
 d_gen(Tag,bot,IqsIn).

dgen(Tag,SVs,Iqs) :-
  retractall(d_mode(_)),
  assert(d_mode(creep)),
  d_gen(Tag,SVs,Iqs).
  
d_gen(Tag,SVs,Iqs) :-
  nl, write('QUERY: '), nl, ttyflush,
  pp_fs(Tag-SVs,Iqs), nl,
  d_gen(Tag,SVs,Iqs,Words),
  nl, write('STRING: '),
  nl, write_list(Words),
  nl, query_proceed.

dgen(Tag,SVs,IqsIn,Words) :-
  retractall(d_mode(_)),
  assert(d_mode(creep)),
  d_gen(Tag,SVs,IqsIn,Words).

d_gen(Tag,SVs,IqsIn,Words) :-
  ( [no,semantics,specification,found] if_warning_else_fail
      (\+ current_predicate(semantics,semantics(_))),
      !
  ; semantics(Pred), Goal =.. [Pred,_,_],
    [no,Pred,definite,clause,found] if_warning_else_fail
      (\+ current_predicate(if,if(Goal,_))), !
  ; d_generate(Tag,SVs,IqsIn,IqsOut,Words,[],0,query),
    extensionalise(Tag,SVs,IqsOut),
    check_inequal(IqsOut,_)
  ).   

d_generate(Tag,SVs,IqsIn,IqsOut,Words,RestWords,Skip,RootSource) :-
  lsemantics(DirLine,Pred),
  d_call_fail(Skip,NewSkip,sem_index(Tag-SVs,IndexTag-bot,IqsIn,RootSource),
                                     DirLine),
  on_exception(ale_fail,(d_solve_goal(Pred,2,[Tag-SVs,IndexTag-bot],
                                      IqsIn,IqsIndex,NewSkip),
                         d_exit_redo(Skip,sem_index(Tag-SVs,IndexTag-bot,
                                                    IqsIndex,RootSource),
                                     DirLine)),
               fail),
  d_call_fail(Skip,NewSkip2,pivot_tmpl(IndexTag-bot,PivotTag-bot,IqsIndex,
                                       RootSource),
              DirLine),
  on_exception(ale_fail,(d_solve_goal(Pred,2,[PivotTag-bot,IndexTag-bot],
                                      IqsIndex,IqsPivot,NewSkip2),
                         d_exit_redo(Skip,pivot_tmpl(IndexTag-bot,PivotTag-bot,
                                                     IqsPivot,RootSource),
                                     DirLine)),
               fail),
  deref(PivotTag,bot,DPivotTag,DPivotSVs),
  check_inequal(IqsPivot,IqsPivot2),
  d_call_fail(Skip,NewSkip3,pivot_find(DPivotTag-DPivotSVs,IqsPivot2,_,
                                       RootSource),
              none),
  on_exception(ale_fail,(d_non_chain_rule(DPivotTag,DPivotSVs,IqsPivot2,
                                          IqsRule,
                                          LDtrs,PivotSource,NCLine,NewSkip3,
                                          RootSource),
                         check_inequal(IqsRule,IqsChecked),
                         d_exit_redo(Skip,pivot_find(DPivotTag-DPivotSVs,
                                                     IqsChecked,PivotSource,
                                                     RootSource),
                                     none)),
               fail),
  (\+ \+  
  (d_call_fail(Skip,NewSkip4,chain_check(Tag-SVs,DPivotTag-DPivotSVs,
                                         IqsChecked,PivotSource,RootSource,
                                         pivot),
               none),
   on_exception(ale_fail,(current_chain_length(Max),
                          d_chained_nodes(0,Max,DPivotTag,DPivotSVs,Tag,SVs,
                                          IqsChecked,IqsChained,NewSkip4,
                                          PivotSource,RootSource,pivot),
                          d_exit_redo(Skip,chain_check(Tag-SVs,
                                                       DPivotTag-DPivotSVs,
                                                       IqsChained,PivotSource,
                                                       RootSource,pivot),
                                      none)),
                fail))
  ),
  d_call_fail(Skip,NewSkip5,pivot_gen(NewWords,RestNewWords,PivotSource,
                                      RootSource,call),
              NCLine),
  on_exception(ale_fail,(d_gen_dtrs(LDtrs,IqsChecked,IqsDtrs,NewWords,
                                    RestNewWords,NewSkip5,1,_,PivotSource),
                         d_exit_redo(Skip,pivot_gen(NewWords,RestNewWords,
                                                    PivotSource,RootSource,
                                                    exit),
                                     NCLine)),
               fail),
  d_call_fail(Skip,NewSkip6,pivot_conn(Tag-SVs,DPivotTag-DPivotSVs,IqsDtrs,
                                       Words,RestWords,NewWords,RestNewWords,
                                       PivotSource,RootSource,call),
              none),
  on_exception(ale_fail,(current_chain_length(Max),
                         d_conn(0,Max,DPivotTag,DPivotSVs,Tag,SVs,IqsDtrs,
                                IqsOut,NewWords,RestNewWords,Words,RestWords,
                                NewSkip6,PivotSource,RootSource,pivot),
                         d_exit_redo(Skip,pivot_conn(Tag-SVs,
                                                     DPivotTag-DPivotSVs,
                                                     IqsOut,Words,RestWords,
                                                     NewWords,RestNewWords,
                                                     PivotSource,RootSource,
                                                     exit),
                                     none)),
               fail).

d_non_chain_rule(PivotTag,PivotSVs,IqsIn,IqsOut,(word> Word),PivotSource,
                 WordLine,Skip,RootSource) :-
  'l--->'(WordStart,WordLine,LabelledDescStart),
  d_call_fail(Skip,NewSkip,pivot_lex(WordStart,PivotTag-PivotSVs,IqsIn,
                                     RootSource),
              WordLine),
  on_exception(ale_fail,(d_add_to(LabelledDescStart,Tag,bot,IqsIn,IqsMid,
                                  NewSkip,lex),
                         curr_lex_rule_depth(Max),
                         d_gen_lex_close(0,Max,WordStart,Tag,bot,Word,
                                         PivotTag,PivotSVs,IqsMid,IqsOut,
                                         PivotSource,NewSkip,RootSource),
                         d_exit_redo(Skip,pivot_lex(WordStart,
                                                    PivotTag-PivotSVs,IqsOut,
                                                    RootSource),
                                     WordLine)),
               fail).
d_non_chain_rule(PivotTag,PivotSVs,IqsIn,IqsOut,empty,empty,OpLine,Skip,
                 RootSource) :-
  current_predicate(lempty,lempty(_,_)),
  lempty(OpLine,LDesc),
  d_call_fail(Skip,NewSkip,pivot_empty(PivotTag-PivotSVs,IqsIn,RootSource),
              OpLine),
  on_exception(ale_fail,(d_add_to(LDesc,PivotTag,PivotSVs,IqsIn,IqsOut,NewSkip,
                                  pivot),
                         d_exit_redo(Skip,pivot_empty(PivotTag-PivotSVs,
                                                      IqsOut,RootSource),
                                     OpLine)),
               fail).
d_non_chain_rule(PivotTag,PivotSVs,IqsIn,IqsOut,LabelledRuleBody,
                 ncrule(RuleName),RuleLine,Skip,RootSource) :-
  current_predicate(lrule,lrule(_,_,_,_)),
  lrule(RuleName,RuleLine,LabelledMother,LabelledRuleBody),
  \+ d_chain_rule_body(LabelledRuleBody),
  d_call_fail(Skip,NewSkip,pivot_ncrule(RuleName,PivotTag-PivotSVs,IqsIn,
                                        RootSource),
              RuleLine),
  on_exception(ale_fail,(d_add_to(LabelledMother,PivotTag,PivotSVs,IqsIn,
                                  IqsOut,NewSkip,pivot),
                         d_exit_redo(Skip,pivot_ncrule(RuleName,
                                                       PivotTag-PivotSVs,
                                                       IqsOut,RootSource),
                                     RuleLine)),
               fail).

d_gen_lex_close(N,_,Word,Tag,SVs,Word,PivotTag,PivotSVs,IqsIn,IqsOut,
                PivotSource,Skip,RootSource) :-
  d_call_fail(Skip,_,lex_pivot_unify(Word,Tag-SVs,PivotTag-PivotSVs,IqsIn,
                                     RootSource),
              none),
  on_exception(ale_fail,(ud(Tag,SVs,PivotTag,PivotSVs,IqsIn,IqsOut),
                         d_exit_redo(Skip,
                                     lex_pivot_unify(Word,Tag-SVs,
                                                     PivotTag-PivotSVs,IqsOut,
                                                     RootSource),
                                     none)),
               fail),
  (N =:= 0
  -> PivotSource = base(Word)
   ; PivotSource = deriv(Word,_)).
d_gen_lex_close(N,Max,WordIn,Tag,SVs,WordOut,PivotTag,PivotSVs,IqsIn,IqsOut,
                PivotSource,Skip,RootSource) :-
  current_predicate(llr,llr(_,_,_)),
  N < Max,
  d_lex_rule(WordIn,Tag,SVs,WordMid,TagMid,SVsMid,IqsIn,IqsMid,Skip),
  NewN is N + 1,
  (N =:= 0
  -> PivotSource = deriv(_,WordIn)
   ; true),
  d_gen_lex_close(NewN,Max,WordMid,TagMid,SVsMid,WordOut,PivotTag,PivotSVs,
                  IqsMid,IqsOut,PivotSource,Skip,RootSource).

d_chained_nodes(_,_,ChainTag,ChainSVs,RootTag,RootSVs,IqsIn,IqsOut,Skip,
                PivotSource,RootSource,ChainSource) :-
  d_call_fail(Skip,_,root_chain_unify(RootTag-RootSVs,ChainTag-ChainSVs,IqsIn,
                                      PivotSource,RootSource,ChainSource),
              none),
  on_exception(ale_fail,(ud(ChainTag,ChainSVs,RootTag,RootSVs,IqsIn,IqsOut),
                         d_exit_redo(Skip,root_chain_unify(RootTag-RootSVs,
                                                           ChainTag-ChainSVs,
                                                           IqsOut,PivotSource,
                                                           RootSource,
                                                           ChainSource),
                                     none)),
               fail).
d_chained_nodes(N,Max,ChainTag,ChainSVs,RootTag,RootSVs,IqsIn,IqsOut,Skip,
                PivotSource,RootSource,ChainSource) :-
  current_predicate(lrule,lrule(_,_,_,_)),
  N < Max,
  lrule(RuleName,RuleLine,LabelledMother,LabelledRuleBody),
  d_chain_rule_body(LabelledRuleBody,_,LSemHead,_),
  d_call_fail(Skip,NewSkip,chain_crule(RuleName,RootTag-RootSVs,
                                       ChainTag-ChainSVs,IqsIn,PivotSource,
                                       RootSource,ChainSource),
              RuleLine),
  on_exception(ale_fail,(
    deref(ChainTag,ChainSVs,DChainTag,DChainSVs),
    d_add_to(LSemHead,DChainTag,DChainSVs,IqsIn,IqsMid,NewSkip,pivot),
    check_inequal(IqsMid,IqsMid2),
    d_add_to(LabelledMother,NewChainTag,bot,IqsMid2,IqsMid3,NewSkip,mother),
    check_inequal(IqsMid3,IqsMid4),
    d_exit_redo(Skip,chain_crule(RuleName,RootTag-RootSVs,
                                 DChainTag-DChainSVs,IqsMid4,PivotSource,
                                 RootSource,ChainSource),
                RuleLine)),
    fail),
  NewN is N + 1,
  d_call_fail(Skip,NewSkip2,chain_check(RootTag-RootSVs,NewChainTag-bot,
                                        IqsMid4,PivotSource,RootSource,
                                        crule(RuleName)),
              none),
  on_exception(ale_fail,(d_chained_nodes(NewN,Max,NewChainTag,bot,RootTag,
                                         RootSVs,IqsMid4,IqsOut,NewSkip2,
                                         PivotSource,RootSource,
                                         crule(RuleName)),
                         d_exit_redo(Skip,chain_check(RootTag-RootSVs,
                                                      NewChainTag-bot,IqsOut,
                                                      PivotSource,RootSource,
                                                      crule(RuleName)),
                                     none)),
               fail).

d_gen_dtrs(empty,Iqs,Iqs,Words,Words,_Skip,N,N,_) :- !.
d_gen_dtrs((word> Word),Iqs,Iqs,[Word|RestWords],RestWords,_Skip,N,SN,_) :-
  !,SN is N + 1.
d_gen_dtrs(lcat(LDtr),IqsIn,IqsOut,Words,RestWords,Skip,N,SN,
           RuleSource) :-
  !,d_add_to(LDtr,DtrTag,bot,IqsIn,IqsDtr,Skip,nonheaddtr(N,RuleSource)),
  deref(DtrTag,bot,NewTag,NewSVs),
  check_inequal(IqsDtr,IqsDtrCkd),
  d_generate(NewTag,NewSVs,IqsDtrCkd,IqsOut,Words,RestWords,Skip,
             nonheaddtr(N,RuleSource)),
  SN is N + 1.
d_gen_dtrs((lcat(LDtr),RestDtrs),IqsIn,IqsOut,Words,RestWords,Skip,N,NewN,
           RuleSource) :-
  !,d_add_to(LDtr,DtrTag,bot,IqsIn,IqsDtr,Skip,nonheaddtr(N,RuleSource)),
  deref(DtrTag,bot,NewTag,NewSVs),
  check_inequal(IqsDtr,IqsDtrCkd),
  d_generate(NewTag,NewSVs,IqsDtrCkd,IqsGen,Words,MidWords,Skip,
             nonheaddtr(N,RuleSource)),
  SN is N + 1,
  d_gen_dtrs(RestDtrs,IqsGen,IqsOut,MidWords,RestWords,Skip,SN,NewN,
             RuleSource).
d_gen_dtrs(lgoal(LGoal),IqsIn,IqsOut,Words,Words,Skip,N,N,_) :-
  !,d_solve(LGoal,IqsIn,IqsOut,Skip).
d_gen_dtrs((lgoal(LGoal),RestDtrs),IqsIn,IqsOut,Words,RestWords,Skip,N,NewN,
           RuleSource) :-
  !,d_solve(LGoal,IqsIn,IqsBody,Skip),
  d_gen_dtrs(RestDtrs,IqsBody,IqsOut,Words,RestWords,Skip,N,NewN,RuleSource).
d_gen_dtrs(lcats(LDtrsList),IqsIn,IqsOut,Words,RestWords,Skip,N,NewN,
           RuleSource) :-
  !,d_add_to(LDtrsList,Tag,bot,IqsIn,IqsList,Skip,dtrlist),
  check_inequal(IqsList,IqsListCkd),
  deref(Tag,bot,_,DSVs), DSVs =.. [Sort|Vs],
  d_gen_list_dtrs(Sort,Vs,IqsListCkd,IqsOut,Words,RestWords,Skip,N,NewN,
                  RuleSource).
d_gen_dtrs((lcats(LDtrsList),RestDtrs),IqsIn,IqsOut,Words,RestWords,Skip,N,
           NewN,RuleSource) :-
  !,d_add_to(LDtrsList,Tag,bot,IqsIn,IqsList,Skip,dtrlist),
  check_inequal(IqsList,IqsListCkd),
  deref(Tag,bot,_,DSVs), DSVs =.. [Sort|Vs],
  d_gen_list_dtrs(Sort,Vs,IqsListCkd,IqsGen,Words,MidWords,Skip,N,NextN,
                  RuleSource),
  d_gen_dtrs(RestDtrs,IqsGen,IqsOut,MidWords,RestWords,Skip,NextN,NewN,
             RuleSource).

d_gen_list_dtrs(e_list,_,Iqs,Iqs,Words,Words,_Skip,N,N,_) :-
  !.
d_gen_list_dtrs(Sort,[HdFS,TlFS],IqsIn,IqsOut,Words,RestWords,Skip,N,NewN,
                RuleSource) :-
  sub_type(ne_list,Sort),
  !,deref(HdFS,DtrTag,DtrSVs),
  d_generate(DtrTag,DtrSVs,IqsIn,IqsDtr,Words,MidWords,Skip,
             nonheaddtr(N,RuleSource)),
  deref(TlFS,_,TlSVs), TlSVs =.. [TlSort|TlVs],
  SN is N + 1,
  d_gen_list_dtrs(TlSort,TlVs,IqsDtr,IqsOut,MidWords,RestWords,Skip,SN,NewN,
                  RuleSource).

d_conn(_,_,PivotTag,PivotSVs,RootTag,RootSVs,IqsIn,IqsOut,Words,RestWords,
       Words,RestWords,Skip,PivotSource,RootSource,ChainSource) :-
  d_call_fail(Skip,_,root_chain_unify(RootTag-RootSVs,PivotTag-PivotSVs,IqsIn,
                                      PivotSource,RootSource,ChainSource),
              none),
  on_exception(ale_fail,(ud(PivotTag,PivotSVs,RootTag,RootSVs,IqsIn,IqsOut),
                         d_exit_redo(Skip,root_chain_unify(RootTag-RootSVs,
                                                           PivotTag-PivotSVs,
                                                           IqsOut,PivotSource,
                                                           RootSource,
                                                           ChainSource),
                                     none)),
               fail).
d_conn(N,Max,PivotTag,PivotSVs,RootTag,RootSVs,IqsIn,IqsOut,
       HeadWords,RestHeadWords,Words,RestWords,Skip,PivotSource,RootSource,
       ChainSource) :-
  current_predicate(lrule,lrule(_,_,_,_)),
  N < Max,
  lrule(RuleName,RuleLine,LabelledMother,LabelledRuleBody),
  d_chain_rule_body(LabelledRuleBody,LSemGoalBefore,LSemHead,LSemGoalAfter,
                    PrevDtrs,NextDtrs),
  d_call_fail(Skip,NewSkip,chain_crule(RuleName,RootTag-RootSVs,
                                       PivotTag-PivotSVs,IqsIn,PivotSource,
                                       RootSource,ChainSource),
              RuleLine),
  on_exception(ale_fail,(
    (LSemGoalBefore = empty
    -> IqsSGBefore = IqsIn
     ; d_solve(LSemGoalBefore,IqsIn,IqsSGBefore,NewSkip)),
    deref(PivotTag,PivotSVs,DPivotTag,DPivotSVs),
    d_add_to(LSemHead,DPivotTag,DPivotSVs,IqsSGBefore,IqsHead,NewSkip,
             sem_head),
    check_inequal(IqsHead,IqsHeadCkd),
    (LSemGoalAfter = empty
    -> IqsSGAfter = IqsHeadCkd
     ; d_solve(LSemGoalAfter,IqsHeadCkd,IqsSGAfter,NewSkip)),
    d_add_to(LabelledMother,ChainTag,bot,IqsSGAfter,IqsMother,NewSkip,
             mother),
    check_inequal(IqsMother,IqsMotherCkd),
    d_exit_redo(Skip,chain_crule(RuleName,RootTag-RootSVs,PivotTag-PivotSVs,
                                 IqsMotherCkd,PivotSource,RootSource,
                                 ChainSource),
                RuleLine)),
    fail),
  NewN is N + 1,
  (\+ \+  
  (d_call_fail(Skip,NewSkip2,chain_check(RootTag-RootSVs,ChainTag-bot,
                                         IqsMotherCkd,PivotSource,RootSource,
                                         crule(RuleName)),
               none),
   on_exception(ale_fail,(
     d_chained_nodes(NewN,Max,ChainTag,bot,RootTag,RootSVs,IqsMotherCkd,
                     IqsChained,NewSkip2,PivotSource,RootSource,
                     crule(RuleName)),
     d_exit_redo(Skip,chain_check(RootTag-RootSVs,ChainTag-bot,IqsChained,
                                  PivotSource,RootSource,crule(RuleName)),
                 none)),
     fail))
  ),
  d_call_fail(Skip,NewSkip3,pre_chain_gen(NewWords,HeadWords,RestHeadWords,
                                          PivotSource,RootSource,ChainSource,
                                          call),
              RuleLine),
  on_exception(ale_fail,(d_gen_dtrs(PrevDtrs,IqsMotherCkd,IqsDtrsBefore,
                                    NewWords,HeadWords,NewSkip3,1,DtrNum,
                                    crule(RuleName)),
                         d_exit_redo(Skip,pre_chain_gen(NewWords,HeadWords,
                                                        RestHeadWords,
                                                        PivotSource,
                                                        RootSource,
                                                        ChainSource,exit),
                                     RuleLine)),
               fail),
  d_call_fail(Skip,NewSkip4,post_chain_gen(NewWords,HeadWords,RestHeadWords,
                                           RestNewWords,PivotSource,RootSource,
                                           ChainSource,call),
              RuleLine),
  on_exception(ale_fail,(d_gen_dtrs(NextDtrs,IqsDtrsBefore,IqsDtrsAfter,
                                    RestHeadWords,RestNewWords,NewSkip4,
                                    DtrNum,_,crule(RuleName)),
                         d_exit_redo(Skip,post_chain_gen(NewWords,HeadWords,
                                                         RestHeadWords,
                                                         RestNewWords,
                                                         PivotSource,
                                                         RootSource,
                                                         ChainSource,exit),
                                     RuleLine)),
               fail),
  d_call_fail(Skip,NewSkip5,new_chain_conn(RootTag-RootSVs,ChainTag-bot,
                                           IqsDtrsAfter,Words,NewWords,
                                           RestNewWords,RestWords,PivotSource,
                                           RootSource,crule(RuleName),call),
              none),
  on_exception(ale_fail,(d_conn(NewN,Max,ChainTag,bot,RootTag,RootSVs,
                                IqsDtrsAfter,IqsOut,NewWords,RestNewWords,
                                Words,RestWords,NewSkip5,PivotSource,
                                RootSource,crule(RuleName)),
                         d_exit_redo(Skip,new_chain_conn(RootTag-RootSVs,
                                                         ChainTag-bot,
                                                         IqsOut,Words,
                                                         NewWords,
                                                         RestNewWords,
                                                         RestWords,PivotSource,
                                                         RootSource,
                                                         crule(RuleName),
                                                         exit),
                                     none)),
               fail).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Debugger options
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% dclear_bps/0: clear all breakpoints
dclear_bps :-
  retractall(ale_break(_)),
  ale_debugging(AbsFile),
  ale_breakable_lines(BLs),
  emacs_format(user_error,'~n10L{BREAKABLELINES "~a" (',[AbsFile]),
  ale_noemacs -> true ; skip_count_translate(BLs),
  emacs_format(user_error,')}10R~n',[]),flush_output(user_error),
  emacs_format(user_error,'~n10L{BREAKPOINTS "~a" ()}10R~n',[AbsFile]),
  flush_output(user_error).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% dskip/1: automatically skip certain components (lex. rules, empty cats
%  etc.)
dskip(-all) :-
  retractall(d_skip(_)).
dskip(+all) :-
  dskip(+lex),
  dskip(+empty),
  dskip(+rule),
  dskip(+desc).

dskip(-lex) :-
  retractall(d_skip(lex(_,_))).
dskip(+lex) :-
  assert(d_skip(lex(_,_))).

dskip(-empty) :-
  retractall(d_skip(empty)).
dskip(+empty) :-
  assert(d_skip(empty)).

dskip(-rule) :-
  retractall(d_skip(rule_close)).
dskip(+rule) :-
  assert(d_skip(rule_close)).

dskip(-desc) :-
  retractall(d_skip(unify(_,_,_,_,_))),
  retractall(d_skip(eq(_,_,_,_,_))),
  retractall(d_skip(ineq_add(_,_,_))),
  retractall(d_skip(featval(_,_,_,_))),
  retractall(d_skip(macro(_,_,_))),
  retractall(d_skip(type(_,_,_,_))),
  retractall(d_skip(atom(_,_,_,_))),
  retractall(d_skip(fun(_,_,_,_))).
dskip(+desc) :-
  assert(d_skip(unify(_,_,_,_,_))),
  assert(d_skip(eq(_,_,_,_,_))),
  assert(d_skip(ineq_add(_,_,_))),
  assert(d_skip(featval(_,_,_,_))),
  assert(d_skip(macro(_,_,_))),
  assert(d_skip(type(_,_,_,_))),
  assert(d_skip(atom(_,_,_,_))),
  assert(d_skip(fun(_,_,_,_))).

ale_dskip(Command) :-
  current_predicate(d_skip,d_skip(_)),
  d_skip(Command).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% dleash/1:  leashing control on components
dleash(+all) :-
  retractall(ale_no_leash(_)).
dleash(-all) :-
  dleash(-empty),
  dleash(-lex),
  dleash(-desc),
  dleash(-lr),
  dleash(-fun),
  dleash(-rel),
  dleash(-rule),
  dleash(-gen).

dleash(+empty) :-
  retractall(ale_no_leash(empty)).
dleash(-empty) :-
  assert(ale_no_leash(empty)).

dleash(+lex) :-
  retractall(ale_no_leash(lex(_,_))).
dleash(-lex) :-
  assert(ale_no_leash(lex(_,_))).

dleash(+desc) :-
  retractall(ale_no_leash(unify(_,_,_,_,_))),
  retractall(ale_no_leash(eq(_,_,_,_,_))),
  retractall(ale_no_leash(ineq_add(_,_,_))),
  retractall(ale_no_leash(featval(_,_,_,_))),
  retractall(ale_no_leash(macro(_,_,_))),
  retractall(ale_no_leash(type(_,_,_,_))),
  retractall(ale_no_leash(atom(_,_,_,_))),
  retractall(ale_no_leash(fun(_,_,_,_))).
dleash(-desc) :-
  assert(ale_no_leash(unify(_,_,_,_,_))),
  assert(ale_no_leash(eq(_,_,_,_,_))),
  assert(ale_no_leash(ineq_add(_,_,_))),
  assert(ale_no_leash(featval(_,_,_,_))),
  assert(ale_no_leash(macro(_,_,_))),
  assert(ale_no_leash(type(_,_,_,_))),
  assert(ale_no_leash(atom(_,_,_,_))),
  assert(ale_no_leash(fun(_,_,_,_))).

dleash(+lr) :-
  retractall(ale_no_leash(lr(_,_,_,_))),
  retractall(ale_no_leash(morph_input(_))),
  retractall(ale_no_leash(morph_output(_,_))),
  retractall(ale_no_leash(when)),
  retractall(ale_no_leash(lex_pivot_unify(_,_,_,_,_))).
dleash(-lr) :-
  assert(ale_no_leash(lr(_,_,_,_))),
  assert(ale_no_leash(morph_input(_))),
  assert(ale_no_leash(morph_output(_,_))),
  assert(ale_no_leash(when)),
  assert(ale_no_leash(lex_pivot_unify(_,_,_,_,_))).

dleash(+fun) :-
  retractall(ale_no_leash(fun_clause(_,_))).
dleash(-fun) :-
  assert(ale_no_leash(fun_clause(_,_))).

dleash(+rel) :-
  retractall(ale_no_leash(comp_clause(_,_))),
  retractall(ale_no_leash(neg)),
  retractall(ale_no_leash(exteq)),
  retractall(ale_no_leash(prolog(_))),
  retractall(ale_no_leash(comp(_,_))).
dleash(-rel) :-
  assert(ale_no_leash(comp_clause(_,_))),
  assert(ale_no_leash(neg)),
  assert(ale_no_leash(exteq)),
  assert(ale_no_leash(prolog(_))),
  assert(ale_no_leash(comp(_,_))).

dleash(+rule) :-
  retractall(ale_no_leash(rule_close)),
  retractall(ale_no_leash(rule(_))).
dleash(-rule) :-
  assert(ale_no_leash(rule_close)),
  assert(ale_no_leash(rule(_))).

dleash(+gen) :-
  retractall(ale_no_leash(sem_index(_,_,_,_))),
  retractall(ale_no_leash(pivot_tmpl(_,_,_,_))),
  retractall(ale_no_leash(pivot_find(_,_,_,_))),
  retractall(ale_no_leash(pivot_lex(_,_,_,_))),
  retractall(ale_no_leash(pivot_empty(_,_,_))),
  retractall(ale_no_leash(pivot_ncrule(_,_,_,_))),
  retractall(ale_no_leash(chain_crule(_,_,_,_,_,_,_))),
  retractall(ale_no_leash(chain_check(_,_,_,_,_,_))),
  retractall(ale_no_leash(pivot_gen(_,_,_,_,_))),
  retractall(ale_no_leash(pre_chain_gen(_,_,_,_,_,_,_))),
  retractall(ale_no_leash(post_chain_gen(_,_,_,_,_,_,_,_))),
  retractall(ale_no_leash(pivot_conn(_,_,_,_,_,_,_,_,_,_))),
  retractall(ale_no_leash(new_chain_conn(_,_,_,_,_,_,_,_,_,_,_))),
  retractall(ale_no_leash(root_chain_unify(_,_,_,_,_,_))).
dleash(-gen) :-
  assert(ale_no_leash(sem_index(_,_,_,_))),
  assert(ale_no_leash(pivot_tmpl(_,_,_,_))),
  assert(ale_no_leash(pivot_find(_,_,_,_))),
  assert(ale_no_leash(pivot_lex(_,_,_,_))),
  assert(ale_no_leash(pivot_empty(_,_,_))),
  assert(ale_no_leash(pivot_ncrule(_,_,_,_))),
  assert(ale_no_leash(chain_crule(_,_,_,_,_,_,_))),
  assert(ale_no_leash(chain_check(_,_,_,_,_,_))),
  assert(ale_no_leash(pivot_gen(_,_,_,_,_))),
  assert(ale_no_leash(pre_chain_gen(_,_,_,_,_,_,_))),
  assert(ale_no_leash(post_chain_gen(_,_,_,_,_,_,_,_))),
  assert(ale_no_leash(pivot_conn(_,_,_,_,_,_,_,_,_,_))),
  assert(ale_no_leash(new_chain_conn(_,_,_,_,_,_,_,_,_,_,_))),
  assert(ale_no_leash(root_chain_unify(_,_,_,_,_,_))).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Call and Fail ports
d_call_fail(Skip,NewSkip,Command,Line) :-
  d_mode(Mode),
  display_and_ask_call(Mode,Command,Line,Skip),
  NewSkip is Skip + 1.
d_call_fail(Skip,NewSkip,Command,Line) :-
  d_mode(Mode),
  display_and_ask_fail(Mode,Skip,Command,Line,NewSkip).

display_and_ask_call(skip(_),_,_,_).
display_and_ask_call(leap,Command,Line,Skip) :-
  ((current_predicate(ale_break,ale_break(_)),ale_break(Line)) ->
    retract(d_mode(_)),
    assert(d_mode(creep)),
    d_ask(call,Command,Line),
    (ale_dskip(Command) ->
      write('<auto-skip>'),nl,ttyflush,
      retract(d_mode(_)),
      assert(d_mode(skip(Skip)))
    ; (ale_leash(Command) ->
        ttyget0(Reply),
        dac_act(Reply,Skip,Command,Line)
      ; nl,ttyflush,
        true))
  ; true).
display_and_ask_call(line(LeapLine),Command,Line,Skip) :-
  ((LeapLine == Line) ->
    true
  ; retract(d_mode(_)),
    assert(d_mode(creep)),
    d_ask(call,Command,Line),
    (ale_dskip(Command) ->
      write('<auto-skip>'),nl,ttyflush,
      retract(d_mode(_)),
      assert(d_mode(skip(Skip)))
    ; (ale_leash(Command) ->
        ttyget0(Reply),
        dac_act(Reply,Skip,Command,Line)
      ; nl,ttyflush,
        true))).
display_and_ask_call(creep,Command,Line,Skip) :-
  d_ask(call,Command,Line),
  (ale_dskip(Command) ->
    write('<auto-skip>'),nl,ttyflush,
    retract(d_mode(_)),
    assert(d_mode(skip(Skip)))
  ; (ale_leash(Command) ->
      ttyget0(Reply),
      dac_act(Reply,Skip,Command,Line)
    ; nl,ttyflush,
      true)).

dac_act(97,_,_,_) :-       % (a)bort
  skip_line,
  abort.
dac_act(102,_,_,_) :-      % (f)ail
  !,skip_line,
  fail.
dac_act(108,_,_,_) :- % (l)eap
  !,skip_line,
  retract(d_mode(_)),
  assert(d_mode(leap)).
dac_act(99,_,_,_) :-  % (c)reep
  !,skip_line.
dac_act(115,Skip,_,_) :-   % (s)kip
  !,skip_line,
  retract(d_mode(_)),
  assert(d_mode(skip(Skip))).
dac_act(110,_,_,Line) :-   % li(n)e
  !,skip_line,
  retract(d_mode(_)),
  assert(d_mode(line(Line))). 
dac_act(100,Skip,Command,Line) :-  % (d)isplay current FS
  !,skip_line,
  display_current(Command),
  display_and_ask_call(creep,Command,Line,Skip).
dac_act(43,Skip,Command,Line) :-        % +: set spypoint
  !,skip_line,
  ale_debugging(File),
  ((current_predicate(ale_break,ale_break(_)),ale_break(Line)) ->
    format(user_error, '{Breakpoint already set}~n', []),
    flush_output(user_error)
  ;(Line == none ->
     format(user_error, '{Could not set breakpoint}~n', []),
     flush_output(user_error)
   ; assert(ale_break(Line)),
     LineFrom1 is Line + 1,
     emacs_format(user_error,'10Lignore10R~n10L{BREAK ~w "~a"}10R~n',
            [LineFrom1,File]),
     flush_output(user_error),
     nl,ttyflush)),
  display_and_ask_call(creep,Command,Line,Skip).
dac_act(45,Skip,Command,Line) :-        % -: remove spypoint
  !,skip_line,
  ale_debugging(File),
  (retract(ale_break(Line)) ->
     LineFrom1 is Line + 1,
     emacs_format(user_error,'10Lignore10R~n10L{REMOVEBREAK ~w "~a"}10R~n',
            [LineFrom1,File]),
     flush_output(user_error),
     nl,ttyflush
   ; format(user_error, '{No breakpoint at that line}~n', []),
     flush_output(user_error)),
  display_and_ask_call(creep,Command,Line,Skip).
dac_act(105,Skip,Command,Line) :-      % toggle chart (i)nterpreter
  !,skip_line,
  (no_interpreter ->
    interp
  ; nointerp),
  ttyflush,
  display_and_ask_call(creep,Command,Line,Skip).
dac_act(64,Skip,Command,Line) :-       % @: command from user/emacs
  !,read(Goal),
  if(call(Goal),
     true,
     (write(no),nl)),  % don't let Goal failure cause failure in trace
  nl,nl,ttyflush,
  write('Call:'),write_command(Command),write('? '),ttyflush,
  ttyget0(Reply),
  dac_act(Reply,Skip,Command,Line).
dac_act(10,_,_,_) :-  % LFD (same as creep)
  !.
dac_act(104,Skip,Command,Line) :-       % (h)elp
  !,skip_line,
  nl,write('Call port commands:'),nl,
  write('(a)bort     (l)eap'),nl,
  write('(LF/c)reep  li(n)e'),nl,
  write('(f)ail      (s)kip'),nl,
  write(' toggle chart (i)nterpreter'),nl,
  write(' (d)isplay current structure'),nl,
  write(' @ <prolog_goal>.: pass goal to Prolog'),nl,
  write(' +: set breakpoint'),nl,
  write(' -: clear breakpoint'),nl,
  write('(?/h)elp'),nl,
  ttyflush,
  display_and_ask_call(creep,Command,Line,Skip).
dac_act(63,Skip,Command,Line) :-        % ?: help also
  !,dac_act(104,Skip,Command,Line).
dac_act(_,Skip,Command,Line) :-     % everything else - try again
  skip_line,
  write(user_error,'Invalid Response (? or h for help)'),nl(user_error),
  flush_output(user_error),
  display_and_ask_call(creep,Command,Line,Skip).

display_and_ask_fail(skip(Skip),Skip,Command,Line,NewSkip) :-
  !,retract(d_mode(_)),
  assert(d_mode(creep)),
  d_ask(fail,Command,Line),
  (ale_dskip(Command) ->
    write('<auto-skipped>'),nl,ttyflush,
    fail
  ; (ale_leash(Command) ->
      ttyget0(Reply),
      daf_act(Reply,Skip,Command,Line,NewSkip)
    ; nl,ttyflush,
      fail)).
display_and_ask_fail(skip(_),_,_,_,_) :-
  fail.
display_and_ask_fail(leap,Skip,Command,Line,NewSkip) :-
  ((current_predicate(ale_break,ale_break(_)),ale_break(Line)) ->
    retract(d_mode(_)),
    assert(d_mode(creep)),
    d_ask(fail,Command,Line),
    (ale_leash(Command) ->
      ttyget0(Reply),
      daf_act(Reply,Skip,Command,Line,NewSkip)
    ; nl,ttyflush,
      fail)
  ; fail).
display_and_ask_fail(line(LeapLine),Skip,Command,Line,NewSkip) :-
  ((LeapLine == Line) ->
    fail
  ; retract(d_mode(_)),
    assert(d_mode(creep)),
    d_ask(fail,Command,Line),
    (ale_leash(Command) ->
      ttyget0(Reply),
      daf_act(Reply,Skip,Command,Line,NewSkip)
    ; nl,ttyflush,
      fail)).
display_and_ask_fail(creep,Skip,Command,Line,NewSkip) :-
  d_ask(fail,Command,Line),
  (ale_leash(Command) ->
    ttyget0(Reply),
    daf_act(Reply,Skip,Command,Line,NewSkip)
  ; nl,ttyflush,
    fail).

daf_act(97,_,_,_,_) :-       % (a)bort
  skip_line,
  abort.
daf_act(108,_,_,_,_) :- % (l)eap
  !,skip_line,
  retract(d_mode(_)),
  assert(d_mode(leap)),
  fail.
daf_act(99,_,_,_,_) :-  % (c)reep
  !,skip_line,
  fail.
daf_act(114,Skip,Command,Line,NewSkip) :- % (r)etry
  !,skip_line,
  d_call_fail(Skip,NewSkip,Command,Line).
daf_act(10,_,_,_,_) :-                    % LFD (same as creep)
  !,fail.
daf_act(110,_,_,Line,_) :-                % li(n)e
  !,skip_line,
  retract(d_mode(_)),
  assert(d_mode(line(Line))),
  fail.
daf_act(100,Skip,Command,Line,NewSkip) :-  % (d)isplay current FS
  !,skip_line,
  display_current(Command),
  display_and_ask_fail(creep,Skip,Command,Line,NewSkip).
daf_act(43,Skip,Command,Line,NewSkip) :-    % +: set spypoint
  !,skip_line,
  ale_debugging(File),
  ((current_predicate(ale_break,ale_break(_)),ale_break(Line)) ->
    format(user_error, '{Breakpoint already set}~n', []),
    flush_output(user_error)
  ;(Line == none ->
     format(user_error, '{Could not set breakpoint}~n', []),
     flush_output(user_error)
   ; assert(ale_break(Line)),
     LineFrom1 is Line + 1,
     emacs_format(user_error,'10Lignore10R~n10L{BREAK ~w "~a"}10R~n',
            [LineFrom1,File]),
     flush_output(user_error),
     nl,ttyflush)),
  display_and_ask_fail(creep,Skip,Command,Line,NewSkip).
daf_act(45,Skip,Command,Line,NewSkip) :-    % -: remove spypoint
  !,skip_line,
  ale_debugging(File),
  (retract(ale_break(Line)) ->
     LineFrom1 is Line + 1,
     emacs_format(user_error,'10Lignore10R~n10L{REMOVEBREAK ~w "~a"}10R~n',
            [LineFrom1,File]),
     flush_output(user_error),
     nl,ttyflush
   ; format(user_error, '{No breakpoint at that line}~n', []),
     flush_output(user_error)),
  display_and_ask_fail(creep,Skip,Command,Line,NewSkip).
daf_act(105,Skip,Command,Line,NewSkip) :-      % toggle chart (i)nterpreter
  !,skip_line,
  (no_interpreter ->
    interp
  ; nointerp),
  ttyflush,
  display_and_ask_fail(creep,Skip,Command,Line,NewSkip).
daf_act(64,Skip,Command,Line,NewSkip) :-       % @: command from user/emacs
  !,read(Goal),
  if(call(Goal),
     true,
     (write(no),nl)),  % don't let Goal failure cause failure in trace
  nl,nl,ttyflush,
  write('Fail:'),write_command(Command),write('? '),ttyflush,
  ttyget0(Reply),
  daf_act(Reply,Skip,Command,Line,NewSkip).
daf_act(104,Skip,Command,Line,NewSkip) :-       % (h)elp
  !,skip_line,
  nl,write('Fail port commands:'),nl,
  write('(a)bort     (l)eap'),nl,
  write('(LF/c)reep  li(n)e'),nl,
  write('            (r)etry'),nl,
  write(' toggle chart (i)nterpreter'),nl,
  write(' (d)isplay current structure'),nl,
  write(' @ <prolog_goal>.: pass goal to Prolog'),nl,
  write(' +: set breakpoint'),nl,
  write(' -: clear breakpoint'),nl,
  write('(?/h)elp'),nl,
  ttyflush,
  display_and_ask_fail(creep,Skip,Command,Line,NewSkip).
daf_act(63,Skip,Command,Line,NewSkip) :-        % ?: help also
  !,daf_act(104,Skip,Command,Line,NewSkip).
daf_act(_,Skip,Command,Line,NewSkip) :-     % everything else - try again
  skip_line,
  write(user_error,'Invalid Response (? or h for help)'),nl(user_error),
  flush_output(user_error),
  display_and_ask_fail(creep,Skip,Command,Line,NewSkip).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Exit and Redo ports
d_exit_redo(Skip,Command,Line) :-
  d_mode(Mode),
  display_and_ask_exit(Mode,Skip,Command,Line).
d_exit_redo(Skip,Command,Line) :-
  d_mode(Mode),
  display_and_ask_redo(Mode,Command,Line,Skip).

display_and_ask_exit(skip(Skip),Skip,Command,Line) :-
  !,retract(d_mode(_)),
  assert(d_mode(creep)),
  d_ask(exit,Command,Line),
  (ale_dskip(Command) ->
    write('<auto-skipped>'),nl,ttyflush
  ; (ale_leash(Command) ->
      ttyget0(Reply),
      dae_act(Reply,Skip,Command,Line)
    ; nl,ttyflush,
      true)).
display_and_ask_exit(skip(_),_,_,_).
display_and_ask_exit(leap,Skip,Command,Line) :-
  ((current_predicate(ale_break,ale_break(_)),ale_break(Line)) ->
    retract(d_mode(_)),
    assert(d_mode(creep)),
    d_ask(exit,Command,Line),
    (ale_leash(Command) ->
      ttyget0(Reply),
      dae_act(Reply,Skip,Command,Line)
    ; nl,ttyflush,
      true)
  ; true).
display_and_ask_exit(line(LeapLine),Skip,Command,Line) :-
  ((LeapLine == Line) ->
    true
  ; retract(d_mode(_)),
    assert(d_mode(creep)),
    d_ask(exit,Command,Line),
    (ale_leash(Command) ->
      ttyget0(Reply),
      dae_act(Reply,Skip,Command,Line)
    ; nl,ttyflush,
      true)).
display_and_ask_exit(creep,Skip,Command,Line) :-
  d_ask(exit,Command,Line),
  (ale_leash(Command) ->
    ttyget0(Reply),
    dae_act(Reply,Skip,Command,Line)
  ; nl,ttyflush,
    true).

dae_act(97,_,_,_) :-       % (a)bort
  skip_line,
  abort.
dae_act(102,_,_,_) :-      % (f)ail
  !,skip_line,
  raise_exception(ale_fail).
dae_act(108,_,_,_) :-   % (l)eap
  !,skip_line,
  retract(d_mode(_)),
  assert(d_mode(leap)).
dae_act(99,_,_,_) :-       % (c)reep
  !,skip_line.
dae_act(10,_,_,_) :-       % LFD (same as creep)
  !.
dae_act(110,_,_,Line) :-   % li(n)e
  !,skip_line,
  retract(d_mode(_)),
  assert(d_mode(line(Line))).
dae_act(100,Skip,Command,Line) :-   % (d)isplay current FS
  !,skip_line,
  display_current(Command),
  display_and_ask_exit(creep,Skip,Command,Line).
dae_act(43,Skip,Command,Line) :-    % +: set spypoint
  !,skip_line,
  ale_debugging(File),
  ((current_predicate(ale_break,ale_break(_)),ale_break(Line)) ->
    format(user_error, '{Breakpoint already set}~n', []),
    flush_output(user_error)
  ;(Line == none ->
     format(user_error, '{Could not set breakpoint}~n', []),
     flush_output(user_error)
   ; assert(ale_break(Line)),
     LineFrom1 is Line + 1,
     emacs_format(user_error,'10Lignore10R~n10L{BREAK ~w "~a"}10R~n',
            [LineFrom1,File]),
     flush_output(user_error),
     nl,ttyflush)),
  display_and_ask_exit(creep,Skip,Command,Line).
dae_act(45,Skip,Command,Line) :-    % -: remove spypoint
  !,skip_line,
  ale_debugging(File),
  (retract(ale_break(Line)) ->
     LineFrom1 is Line + 1,
     emacs_format(user_error,'10Lignore10R~n10L{REMOVEBREAK ~w "~a"}10R~n',
            [LineFrom1,File]),
     flush_output(user_error),
     nl,ttyflush
   ; format(user_error, '{No breakpoint at that line}~n', []),
     flush_output(user_error)),
  display_and_ask_exit(creep,Skip,Command,Line).
dae_act(105,Skip,Command,Line) :-      % toggle chart (i)nterpreter
  !,skip_line,
  (no_interpreter ->
    interp
  ; nointerp),
  ttyflush,
  display_and_ask_exit(creep,Skip,Command,Line).
dae_act(64,Skip,Command,Line) :-       % @: command from user/emacs
  !,read(Goal),
  if(call(Goal),
     true,
     (write(no),nl)),  % don't let Goal failure cause failure in trace
  nl,nl,ttyflush,
  write('Exit:'),write_command(Command),write('? '),ttyflush,
  ttyget0(Reply),
  dae_act(Reply,Skip,Command,Line).
dae_act(104,Skip,Command,Line) :-       % (h)elp
  !,skip_line,
  nl,write('Exit port commands:'),nl,
  write('(a)bort     (l)eap'),nl,
  write('(LF/c)reep  li(n)e'),nl,
  write('(f)ail'),nl,
  write(' toggle chart (i)nterpreter'),nl,
  write(' (d)isplay current structure'),nl,
  write(' @ <prolog_goal>.: pass goal to Prolog'),nl,
  write(' +: set breakpoint'),nl,
  write(' -: clear breakpoint'),nl,
  write('(?/h)elp'),nl,
  ttyflush,
  display_and_ask_exit(creep,Skip,Command,Line).
dae_act(63,Skip,Command,Line) :-        % ?: help also
  !,dae_act(104,Skip,Command,Line).
dae_act(_,Skip,Command,Line) :-   % everything else - try again
  skip_line,
  write(user_error,'Invalid Response (? or h for help)'),nl(user_error),
  flush_output(user_error),
  display_and_ask_exit(creep,Skip,Command,Line).

display_and_ask_redo(skip(_),_,_,_) :-
  fail.
display_and_ask_redo(leap,Command,Line,Skip) :-
  ((current_predicate(ale_break,ale_break(_)),ale_break(Line)) ->
    retract(d_mode(_)),
    assert(d_mode(creep)),
    d_ask(redo,Command,Line),
    (ale_dskip(Command) ->
      write('<auto-skip>'),nl,ttyflush,
      retract(d_mode(_)),
      assert(d_mode(skip(Skip))),
      fail
    ; (ale_leash(Command) ->
        ttyget0(Reply),
        dar_act(Reply,Skip,Command,Line)
      ; nl,ttyflush,
        fail))
  ; fail).
display_and_ask_redo(line(LeapLine),Command,Line,Skip) :-
  ((LeapLine == Line) ->
    fail
  ; retract(d_mode(_)),
    assert(d_mode(creep)),
    d_ask(redo,Command,Line),
    (ale_dskip(Command) ->
      write('<auto-skip>'),nl,ttyflush,
      retract(d_mode(_)),
      assert(d_mode(skip(Skip))),
      fail
    ; (ale_leash(Command) ->
        ttyget0(Reply),
        dar_act(Reply,Skip,Command,Line)
      ; nl,ttyflush,
        fail))).
display_and_ask_redo(creep,Command,Line,Skip) :-
  d_ask(redo,Command,Line),
  (ale_dskip(Command) ->
    write('<auto-skip>'),nl,ttyflush,
    retract(d_mode(_)),
    assert(d_mode(skip(Skip))),
    fail
  ; (ale_leash(Command) ->
      ttyget0(Reply),
      dar_act(Reply,Skip,Command,Line)
    ; nl,ttyflush,
      fail)).

dar_act(97,_,_,_) :-    % (a)bort
  skip_line,
  abort.
dar_act(102,_,_,_) :-   % (f)ail
  !,skip_line,
  raise_exception(ale_fail).
dar_act(108,_,_,_) :-   % (l)eap
  !,skip_line,
  retract(d_mode(_)),
  assert(d_mode(leap)),
  fail.
dar_act(99,_,_,_) :-       % (c)reep
  !,skip_line,
  fail.
dar_act(115,Skip,_,_) :-   % (s)kip
  !,skip_line,
  retract(d_mode(_)),
  assert(d_mode(skip(Skip))),
  fail.
dar_act(10,_,_,_) :-       % LFD (same as creep)
  !,fail.
dar_act(110,_,_,Line) :-   % li(n)e
  !,skip_line,
  retract(d_mode(_)),
  assert(d_mode(line(Line))),
  fail.
dar_act(100,Skip,Command,Line) :-   % (d)isplay current FS
  !,skip_line,
  display_current(Command),
  display_and_ask_redo(creep,Command,Line,Skip).
dar_act(43,Skip,Command,Line) :-    % +: set spypoint
  !,skip_line,
  ale_debugging(File),
  ((current_predicate(ale_break,ale_break(_)),ale_break(Line)) ->
    format(user_error, '{Breakpoint already set}~n', []),
    flush_output(user_error)
  ;(Line == none ->
     format(user_error, '{Could not set breakpoint}~n', []),
     flush_output(user_error)
   ; assert(ale_break(Line)),
     LineFrom1 is Line + 1,
     emacs_format(user_error,'10Lignore10R~n10L{BREAK ~w "~a"}10R~n',
            [LineFrom1,File]),
     flush_output(user_error),
     nl,ttyflush)),
  display_and_ask_redo(creep,Command,Line,Skip).
dar_act(45,Skip,Command,Line) :-    % -: remove spypoint
  !,skip_line,
  ale_debugging(File),
  (retract(ale_break(Line)) ->
     LineFrom1 is Line + 1,
     emacs_format(user_error,'10Lignore10R~n10L{REMOVEBREAK ~w "~a"}10R~n',
            [LineFrom1,File]),
     flush_output(user_error),
     nl,ttyflush
   ; format(user_error, '{No breakpoint at that line}~n', []),
     flush_output(user_error)),
  display_and_ask_redo(creep,Command,Line,Skip).
dar_act(105,Skip,Command,Line) :-      % toggle chart (i)nterpreter
  !,skip_line,
  (no_interpreter ->
    interp
  ; nointerp),
  ttyflush,
  display_and_ask_redo(creep,Command,Line,Skip).
dar_act(64,Skip,Command,Line) :-       % @: command from user/emacs
  !,read(Goal),
  if(call(Goal),
     true,
     (write(no),nl)),  % don't let Goal failure cause failure in trace
  nl,nl,ttyflush,
  write('Redo:'),write_command(Command),write('? '),ttyflush,
  ttyget0(Reply),
  dar_act(Reply,Skip,Command,Line).
dar_act(104,Skip,Command,Line) :-       % (h)elp
  !,skip_line,
  nl,write('Redo port commands:'),nl,
  write('(a)bort     (l)eap'),nl,
  write('(LF/c)reep  li(n)e'),nl,
  write('(f)ail      (s)kip'),nl,
  write(' toggle chart (i)nterpreter'),nl,
  write(' (d)isplay current structure'),nl,
  write(' @ <prolog_goal>.: pass goal to Prolog'),nl,
  write(' +: set breakpoint'),nl,
  write(' -: clear breakpoint'),nl,
  write('(?/h)elp'),nl,
  ttyflush,
  display_and_ask_redo(creep,Command,Line,Skip).
dar_act(63,Skip,Command,Line) :-        % ?: help also
  !,dar_act(104,Skip,Command,Line).
dar_act(_,Skip,Command,Line) :-    % everything else - try again
  skip_line,
  write(user_error,'Invalid Response (? or h for help)'),nl(user_error),
  flush_output(user_error),
  display_and_ask_redo(creep,Command,Line,Skip).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% d_ask/3: Query user at Port about Command at source-file line Line

d_ask(Port,Command,Line) :-
    ale_debugging(AbsFile),
    ((Line = none) ->
       true
    ; LineFrom1 is Line + 1,
      emacs_format(user_error,'9L~w9M~a9R~n',[LineFrom1,AbsFile]),
      flush_output(user_error)),
    write_port(Port),write_command(Command),write('? '),ttyflush.

write_port(call) :-
  write('Call:').
write_port(fail) :-
  write('Fail:').
write_port(redo) :-
  write('Redo:').
write_port(exit) :-
  write('Exit:').

write_command(unify(Loc,XName,_,_,_)) :-
  write('unify '),write_loc(Loc),write(' with '),write(XName).
write_command(eq(Loc,Path1,Path2,_,_)) :-
  write('path equate '),write(Path1),write(' with '),write(Path2),
  write(' at '),write_loc(Loc).
write_command(ineq_add(Loc,_,_)) :-
  write('inequate '),write_loc(Loc),write(' with description').
write_command(featval(Loc,Feat,_,_)) :-
  write('enforce description on '),write(Feat),write(' value of '),
  write_loc(Loc).
write_command(macro(MacroName,_,_)) :-
  functor(MacroName,Functor,Arity),
  write('substitute macro description for '),write(Functor),write('/'),
  write(Arity).
write_command(type(Loc,Type,_,_)) :-
  write('add type, '),write(Type),write(', to '),write_loc(Loc).
write_command(atom(Loc,Atom,_,_)) :-
  write('add atom, '),write(Atom),write(', to '),write_loc(Loc).
write_command(fun(Functor,Arity,_,_)) :-
  write('evaluate functional description, '),write(Functor),write('/'),
  write(Arity).

write_command(empty) :-
  write('empty category').
write_command(lex(Word,WordStart)) :-
  write('derive: '),write(Word),
  write(' from base entry: '),write(WordStart).
write_command(fun_clause(Functor,Arity)) :-
  write('evaluate functional clause for '),write(Functor),write('/'),
  write(Arity).
write_command(comp_clause(Rel,Arity)) :-
  write('resolve relational clause for '),write(Rel),write('/'),write(Arity).

write_command(lr(RuleName,WordIn,_,_)) :-
  write('apply lexical rule, '),write(RuleName),
  write(', to '),write(WordIn).
write_command(morph_input(WordIn)) :-
  write('apply morph to input: '),write(WordIn).
write_command(morph_output(WordIn,WordOut)) :-
  write('apply morph to input: '),write(WordIn),write(', output: '),
  write(WordOut).
write_command(when) :-
  write('apply morph condition').
write_command(lex_pivot_unify(Word,_,_,_,_)) :-
  write('unify pivot template and '),write(Word).

write_command(neg) :-
  write('resolve negated goal').
write_command(exteq) :-
  write('resolve extensional identity').
write_command(prolog(PGoal)) :-
  write('resolve prolog hook: '),write(PGoal).
write_command(comp(Functor,Arity)) :-
  write('resolve goal, '),write(Functor),write('/'),write(Arity).

write_command(rule_close) :-
  write('close chart edge under rule application').
write_command(rule(RuleName)) :-
  write('apply rule, '),write(RuleName).

write_command(sem_index(_,_,_,_)) :-
  write('build semantic index from root').
write_command(pivot_tmpl(_,_,_,_)) :-
  write('build pivot template from index').
write_command(pivot_find(_,_,_,_)) :-
  write('find pivot').
write_command(pivot_lex(WordStart,_,_,_)) :-
  write('match pivot against entry derived from '),write(WordStart).
write_command(pivot_empty(_,_,_)) :-
  write('match pivot against empty category').
write_command(pivot_ncrule(RuleName,_,_,_)) :-
  write('match pivot against mother of non-chain rule, '),write(RuleName).
write_command(chain_crule(RuleName,_,_,_,_,_,ChainSource)) :-
  write('apply chain rule, '),write(RuleName),write(', to '),
  write_source(ChainSource).
write_command(chain_check(_,_,_,_,_,ChainSource)) :-
  write('check for link from '),write_source(ChainSource),write(' to root').
write_command(pivot_gen(_,_,_,_,_)) :-
  write('recursively generate from pivot').
write_command(pre_chain_gen(_,_,_,_,_,ChainSource,_)) :-
  write('recursively generate pre-head daughters from '),
  write_source(ChainSource).
write_command(post_chain_gen(_,_,_,_,_,_,ChainSource,_)) :-
  write('recursively generate post-head daughters from '),
  write_source(ChainSource).
write_command(pivot_conn(_,_,_,_,_,_,_,_,_,_)) :-
  write('connect pivot to root').
write_command(new_chain_conn(_,_,_,_,_,_,_,_,_,_,_)) :-
  write('connect new chain node to root').
write_command(root_chain_unify(_,_,_,_,_,ChainSource)) :-
  write('unify '),write_source(ChainSource),write(' with root').


write_loc(nonheaddtr(N,RuleSource)) :-
  write('non-head daughter '),write(N),write(', '),
  (RuleSource = ncrule(RuleName)
  -> write('non-chain rule '),write(RuleName)
   ; RuleSource = crule(RuleName),
     write('chain rule '),write(RuleName)).
write_loc(pivot) :-
  write('pivot').
write_loc(empty) :-
  write('empty cat').
write_loc(ineq) :-
  write('inequated desc').
write_loc(feat(F)) :-
  write('value at '),write(F).
write_loc(lex) :-
  write('lex entry').
write_loc(lrin) :-
  write('LR input').
write_loc(lrout) :-
  write('LR output').
write_loc(left) :-
  write('left arg').
write_loc(right) :-
  write('right arg').
write_loc(arg(N)) :-
  write('arg '),write(N).
write_loc(query_desc) :-
  write('query desc').
write_loc(edge) :-
  write('edge').
write_loc(dtrlist) :-
  write('dtr list').
write_loc(mother) :-
  write('mother').
write_loc(sem_head) :-
  write('semantic head').

display_current(unify(_,VarName,FS,Iqs,Var)) :-
  !,(\+ \+ ((var(Var) ->
              Var = _-bot
            ; true),
            duplicates_list([FS,Var],Iqs,[],_,[],_,0,_),
            write('CURRENT STRUCTURE:'),nl,
            pp_fs(FS,[],VisMid,0),nl,
            write('VARIABLE, '),write(VarName),write(':'),nl,
            pp_fs(Var,VisMid,VisMid2,0),nl,
            pp_iqs(Iqs,VisMid2,_,0),nl)).
display_current(eq(_,_,_,FS,Iqs)) :-
  !,pp_fs(FS,Iqs).
display_current(ineq_add(_,FS,Iqs)) :-
  !,pp_fs(FS,Iqs).
display_current(featval(_,_,FS,Iqs)) :-
  !,pp_fs(FS,Iqs).
display_current(macro(_,FS,Iqs)) :-
  !,pp_fs(FS,Iqs).
display_current(type(_,_,FS,Iqs)) :-
  !,pp_fs(FS,Iqs).
display_current(atom(_,_,FS,Iqs)) :-
  !,pp_fs(FS,Iqs).
display_current(fun(_,_,FS,Iqs)) :-
  !,pp_fs(FS,Iqs).
display_current(lr(_,_,FS,Iqs)) :-
  !,pp_fs(FS,Iqs).
display_current(lex_pivot_unify(Word,LexFS,PivotFS,Iqs,RootSource)) :-
  !,write('ROOT: '),write_source(RootSource),nl,
    (\+ \+ (duplicates_list([LexFS,PivotFS],Iqs,[],_,[],_,0,_),
            write('STRUCTURE FOR '),write(Word),write(':'),nl,
            pp_fs(LexFS,[],VisMid,0),nl,
            write('PIVOT:'),nl,
            pp_fs(PivotFS,VisMid,VisMid2,0),nl,
            pp_iqs(Iqs,VisMid2,_,0),nl)).
display_current(sem_index(QFS,IndexFS,Iqs,RootSource)) :-
  !,(\+ \+ (duplicates_list([QFS,IndexFS],Iqs,[],_,[],_,0,_),
            write('ROOT: '),write_source(RootSource),nl,
            pp_fs(QFS,[],VisMid,0),nl,
            write('INDEX:'),nl,
            pp_fs(IndexFS,VisMid,VisMid2,0),nl,
            pp_iqs(Iqs,VisMid2,_,0),nl)).
display_current(pivot_tmpl(IndexFS,PivotFS,Iqs,RootSource)) :-
  !,(\+ \+ (duplicates_list([IndexFS,PivotFS],Iqs,[],_,[],_,0,_),
            write('ROOT: '),write_source(RootSource),nl,
            write('INDEX:'),nl,
            pp_fs(IndexFS,[],VisMid,0),nl,
            write('PIVOT TEMPLATE:'),nl,
            pp_fs(PivotFS,VisMid,VisMid2,0),nl,
            pp_iqs(Iqs,VisMid2,_,0),nl)).
display_current(pivot_find(PivotFS,Iqs,PivotSource,RootSource)) :-
  !,write('ROOT: '),write_source(RootSource),nl,
  write('PIVOT'),
  (nonvar(PivotSource)
  -> write(': '),
     write_source(PivotSource)
   ; write(' TEMPLATE:')),
  nl,
  pp_fs(PivotFS,Iqs).
display_current(pivot_lex(_,PivotFS,Iqs,RootSource)) :-
  !,write('ROOT: '),write_source(RootSource),nl,
  write('PIVOT:'),nl,
  pp_fs(PivotFS,Iqs).
display_current(pivot_empty(PivotFS,Iqs,RootSource)) :-
  !,write('ROOT: '),write_source(RootSource),nl,
  write('PIVOT:'),nl,
  pp_fs(PivotFS,Iqs).
display_current(pivot_ncrule(_,PivotFS,Iqs,RootSource)) :-
  !,write('ROOT: '),write_source(RootSource),nl,
  write('PIVOT:'),nl,
  pp_fs(PivotFS,Iqs).
display_current(chain_crule(_,RootFS,ChainFS,Iqs,PivotSource,RootSource,
                            ChainSource)) :-
  !,(\+ \+ (duplicates_list([RootFS,ChainFS],Iqs,[],_,[],_,0,_),
            write('ROOT: '),write_source(RootSource),nl,
            pp_fs(RootFS,[],VisMid,0),nl,
            write('CHAIN NODE: '),write_source(ChainSource),nl,
            pp_fs(ChainFS,VisMid,VisMid2,0),nl,
            pp_iqs(Iqs,VisMid2,_,0),nl,
            write('PIVOT: '),write_source(PivotSource),nl)).
display_current(chain_check(RootFS,ChainFS,Iqs,PivotSource,RootSource,
                            ChainSource)) :-
  !,(\+ \+ (duplicates_list([RootFS,ChainFS],Iqs,[],_,[],_,0,_),
            write('ROOT: '),write_source(RootSource),nl,
            pp_fs(RootFS,[],VisMid,0),nl,
            write('CHAIN NODE: '),write_source(ChainSource),nl,
            pp_fs(ChainFS,VisMid,VisMid2,0),nl,
            pp_iqs(Iqs,VisMid2,_,0),nl)),
  write('PIVOT: '),write_source(PivotSource),nl.
display_current(pivot_gen(NewWords,RestNewWords,PivotSource,RootSource,
                          Port)) :-
  !,write('ROOT: '),write_source(RootSource),nl,
  write('PIVOT: '),write_source(PivotSource),nl,
  (Port == call
  -> true
   ; write('WORDS: '),nl,
     number_display_head(NewWords,RestNewWords,RestNewWords,'P',nodots,0),nl).
display_current(pre_chain_gen(NewWords,HeadWords,RestHeadWords,
                              PivotSource,RootSource,ChainSource,Port)) :-
  !,write('ROOT: '),write_source(RootSource),nl,
  write('CHAIN NODE: '),write_source(ChainSource),nl,
  write('PIVOT: '),write_source(PivotSource),nl,
  write('WORDS:'),nl,
  (Port == call
  -> write('... '),
     number_display_head(HeadWords,RestHeadWords,RestHeadWords,'C',dots,0)
   ; number_display_pre(NewWords,HeadWords,RestHeadWords,RestHeadWords,'C',
                        dots,0)
  ),
  nl.
display_current(post_chain_gen(NewWords,HeadWords,RestHeadWords,RestNewWords,
                               PivotSource,RootSource,ChainSource,Port)) :-
  !,write('ROOT: '),write_source(RootSource),nl,
  write('CHAIN NODE: '),write_source(ChainSource),nl,
  write('PIVOT: '),write_source(PivotSource),nl,
  write('WORDS:'),nl,
  (Port == call
  -> number_display_pre(NewWords,HeadWords,RestHeadWords,RestHeadWords,'C',
                        dots,0)
   ; number_display_pre(NewWords,HeadWords,RestHeadWords,RestNewWords,'C',
                        nodots,0)
  ),
  nl.
display_current(pivot_conn(RootFS,PivotFS,Iqs,Words,RestWords,HeadWords,
                           RestHeadWords,PivotSource,RootSource,Port)) :-
  !,write('WORDS:'),nl,
  (Port == call
  -> write('... '),
     number_display_head(HeadWords,RestHeadWords,RestHeadWords,'P',dots,0)
   ; number_display_pre(Words,HeadWords,RestHeadWords,RestWords,'P',nodots,0)
  ),
  nl,
  (\+ \+ (duplicates_list([RootFS,PivotFS],Iqs,[],_,[],_,0,_),
          write('ROOT: '),write_source(RootSource),nl,
          pp_fs(RootFS,[],VisMid,0),nl,
          write('PIVOT: '),write_source(PivotSource),nl,
          pp_fs(PivotFS,VisMid,VisMid2,0),nl,
          pp_iqs(Iqs,VisMid2,_,0),nl)).
display_current(new_chain_conn(RootFS,ChainFS,Iqs,Words,HeadWords,
                               RestHeadWords,RestWords,PivotSource,RootSource,
                               ChainSource,Port)) :-
  !,
  write('WORDS:'),nl,
  (Port == call
  -> write('... '),
     number_display_head(HeadWords,RestHeadWords,RestHeadWords,'C',dots,0)
   ; number_display_pre(Words,HeadWords,RestHeadWords,RestWords,'C',nodots,0)
  ),
  nl,
  (\+ \+ (duplicates_list([RootFS,ChainFS],Iqs,[],_,[],_,0,_),
          write('ROOT: '),write_source(RootSource),nl,
          pp_fs(RootFS,[],VisMid,0),nl,
          write('CHAIN NODE: '),write_source(ChainSource),nl,
          pp_fs(ChainFS,VisMid,VisMid2,0),nl,
          pp_iqs(Iqs,VisMid2,_,0),nl,
          write('PIVOT: '),write_source(PivotSource),nl)).
display_current(root_chain_unify(RootFS,ChainFS,Iqs,PivotSource,RootSource,
                                 ChainSource)) :-
  !,(\+ \+ (duplicates_list([RootFS,ChainFS],Iqs,[],_,[],_,0,_),
            write('ROOT: '),write_source(RootSource),nl,
            pp_fs(RootFS,[],VisMid,0),nl,
            write('CHAIN NODE: '),write_source(ChainSource),nl,
            pp_fs(ChainFS,VisMid,VisMid2,0),nl,
            pp_iqs(Iqs,VisMid2,_,0),nl,
            write('PIVOT: '),write_source(PivotSource),nl)).
display_current(_) :-
  write(user_error,'Not valid at this port'),nl(user_error),
  flush_output(user_error).

write_source(base(Word)) :-
  write(Word).
write_source(deriv(Word,WordStart)) :-
  write(Word),write(' (derived from '),write(WordStart),write(')').
write_source(empty) :-
  write('empty category').
write_source(ncrule(RuleName)) :-
  write('mother of non-chain rule '),write(RuleName).
write_source(crule(RuleName)) :-
  write('mother of chain rule '),write(RuleName).
write_source(nonheaddtr(N,RuleSource)) :-
  write('non-head daughter '),write(N),write(', '),
  (RuleSource = ncrule(RuleName)
  -> write('non-chain rule '),write(RuleName)
   ; RuleSource = crule(RuleName),
     write('chain rule '),write(RuleName)).
write_source(pivot) :-
  write('pivot').
write_source(query) :-
  write('query').

number_display_pre(Words,HeadWords,RestHeadWords,RestWords,Delimiter,Dots,N) :-
  Words == HeadWords,
  !,number_display_head(HeadWords,RestHeadWords,RestWords,Delimiter,Dots,N).
number_display_pre([W|Ws],HeadWords,RestHeadWords,RestWords,Delimiter,Dots,
                   N) :-
  write(N),write(' '),write(W),write(' '),
  SN is N + 1,
  number_display_pre(Ws,HeadWords,RestHeadWords,RestWords,Delimiter,Dots,SN).

number_display_head(HeadWords,RestHeadWords,RestWords,Delimiter,Dots,N) :-
  HeadWords == RestHeadWords,  % empty head
  !,write(Delimiter),write(' '),
  number_display_post(RestHeadWords,RestWords,Dots,N).
number_display_head([W|Ws],RestHeadWords,RestWords,Delimiter,Dots,N) :-
  write(Delimiter),write(' '),write(W),write(' '),
  SN is N + 1,
  number_display_head_rest(Ws,RestHeadWords,RestWords,Delimiter,Dots,SN).

number_display_head_rest(HeadWords,RestHeadWords,RestWords,Delimiter,Dots,N) :-
  HeadWords == RestHeadWords,
  !,write(Delimiter),write(' '),
  number_display_post(RestHeadWords,RestWords,Dots,N).
number_display_head_rest([W|Ws],RestHeadWords,RestWords,Delimiter,Dots,N) :-
  write(N),write(' '),write(W),write(' '),
  SN is N + 1,
  number_display_head_rest(Ws,RestHeadWords,RestWords,Delimiter,Dots,SN).

number_display_post(PostWords,RestWords,Dots,_) :-
  PostWords == RestWords,
  !,(Dots = dots
    -> write('...')
     ; true).
number_display_post([W|Ws],RestWords,Dots,N) :-
  SN is N + 1,
  write(W),write(' '),write(SN),write(' '),
  number_display_post(Ws,RestWords,Dots,SN).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Breakpoint glue code for emacs-lisp 

ale_break_at(File,Line) :-
  ale_debugging(File),
  ale_breakable_lines(BLs),
  LineFrom0 is Line - 1,
  member(LineFrom0,BLs),
  !,assert(ale_break(LineFrom0)),
  emacs_format(user_error,'10L{BREAK ~w "~a"}10R',[Line,File]),
  flush_output(user_error).
ale_break_at(_,_) :-
  format(user_error, '~n{Could not set breakpoint}~n', []),
  flush_output(user_error),
  emacs:write_goal_skip(on).

ale_remove_lc(File,Line,Line1) :-
  ale_debugging(File),
  LineFrom0 is Line - 1,
  retract(ale_break(LineFrom0)),
  !,emacs_format(user_error,'10L{REMOVEBREAK ~w "~a"}10R',[Line,File]),
  flush_output(user_error),
  Line = Line1.
ale_remove_lc(_,Line0,Line) :-
  Line = Line0.   % a favour for emacs:emacs_remove_line_condition/2

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% ale_leash/1: check leashing on current command
ale_leash(Command) :-
  current_predicate(ale_no_leash,ale_no_leash(_)),
  ale_no_leash(Command),
  !,fail.
ale_leash(_).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Labelling predicates - intercept token stream directly and record line
%  numbers of source-code

label_items(File) :-
  absolute_file_name(File,AbsFile),
  open(AbsFile,read,GramS),
  label_preds(GramS,AbsFile,BLs,BLs).

label_preds(GramS,AbsFile,BLs,BLsRest) :-
  my_read_tokens(GramS,TokenS,_),
  source_debug:read(TokenS,1200,Term,Left),
  label_term(Term,GramS,AbsFile,TokenS,Left,BLs,BLsRest).

label_term(end_of_file,GramS,AbsFile,_,_,BLs,[]) :-
  close(GramS),
  sort(BLs,BreakableLines),
  assert(ale_debugging(AbsFile)),
  emacs_format(user_error,'~n10L{BREAKABLELINES "~a" (',[AbsFile]),
  (ale_noemacs -> true ; skip_count_translate(BreakableLines)),
  emacs_format(user_error,')}10R~n',[]),flush_output(user_error),
  emacs_format(user_error,'~n10L{BREAKPOINTS "~a" ()}10R~n',[AbsFile]),
  flush_output(user_error),
  assert(ale_breakable_lines(BreakableLines)).
label_term('--->'(Word,Desc),GramS,AbsFile,TokenS,['.'],
           BLs,[WordLine|BLsMid]) :-
  !,trunc_select(atom(Word,WordLine,_),TokenS,TokenSMid),
  trunc_select(atom('--->',_,_),TokenSMid,TokenSRest),
  label_desc(Desc,TokenSRest,LabelledDesc,_,_,BLsMid,BLsRest),
  assertz('l--->'(Word,WordLine,LabelledDesc)),
  label_preds(GramS,AbsFile,BLs,BLsRest).
label_term(if(Head,Body),GramS,AbsFile,TokenS,['.'],BLs,BLsMid) :-
  !,label_compound(Head,TokenS,LabelledHead,_,TokenSMid,BLsMid,BLsMid2),
  trunc_select(atom('if',_,_),TokenSMid,TokenSMid2),
  label_body(Body,TokenSMid2,LabelledBody,_,BLsMid2,BLsRest),
  assertz(lif(LabelledHead,LabelledBody)),
  label_preds(GramS,AbsFile,BLs,BLsRest).
label_term('+++>'(FHead,FResult),GramS,AbsFile,TokenS,['.'],BLs,BLsMid) :-
  !,label_compound(FHead,TokenS,LabelledHead,_,TokenSMid,BLsMid,BLsMid2),
  trunc_select(atom('+++>',_,_),TokenSMid,TokenSMid2),
  label_desc(FResult,TokenSMid2,LabelledResult,_,_,BLsMid2,BLsRest),
  assertz('l+++>'(LabelledHead,LabelledResult)),
  label_preds(GramS,AbsFile,BLs,BLsRest).
label_term(lex_rule(RuleName,LRBody),GramS,AbsFile,TokenS,['.'],BLs,BLsMid) :-
  !,trunc_select(atom(RuleName,_,_),TokenS,TokenSMid),
  trunc_select(atom('lex_rule',_,_),TokenSMid,TokenSMid2),
  label_lr_body(LRBody,TokenSMid2,LabelledLRBody,InLine,BLsMid,BLsRest),
  assertz(llr(RuleName,InLine,LabelledLRBody)),
  label_preds(GramS,AbsFile,BLs,BLsRest).
label_term(rule(RuleName,'===>'(Mother,RuleBody)),GramS,AbsFile,TokenS,['.'],
           BLs,[RuleLine|BLsMid]) :-
  !,trunc_select(atom(RuleName,_,_),TokenS,TokenSMid),
  trunc_select(atom('rule',RuleLine,_),TokenSMid,TokenSMid2),
  label_desc(Mother,TokenSMid2,LabelledMother,_,TokenSMid3,BLsMid,BLsMid2),
  trunc_select(atom('===>',_,_),TokenSMid3,TokenSMid4),
  label_rule_body(RuleBody,TokenSMid4,LabelledRuleBody,_,BLsMid2,BLsRest),
  assertz(lrule(RuleName,RuleLine,LabelledMother,LabelledRuleBody)),
  label_preds(GramS,AbsFile,BLs,BLsRest).
label_term(macro(MacroHead,MacroDesc),GramS,AbsFile,TokenS,['.'],
           BLs,BLsMid):-
  !,strip_immed_left_paren(TokenS,TokenSMid),
  source_debug:read(TokenSMid,1149,MacroHead,TokenSMid2), % (macro=1150)
  trunc_select(atom(macro,_,_),TokenSMid2,TokenSMid3),
  label_desc(MacroDesc,TokenSMid3,LabelledDesc,_,_,BLsMid,BLsRest),
  capture_macro_vars(LabelledDesc,MacroHead,LabelledMacroDesc),
  assertz(lmacro(MacroHead,LabelledMacroDesc)),
  label_preds(GramS,AbsFile,BLs,BLsRest).
label_term((empty Desc),GramS,AbsFile,TokenS,['.'],BLs,[OpLine|BLsMid]) :-
  !,trunc_select(atom('empty',OpLine,_),TokenS,TokenSMid),
  label_desc(Desc,TokenSMid,LabelledDesc,_,_,BLsMid,BLsRest),
  assertz(lempty(OpLine,LabelledDesc)),
  label_preds(GramS,AbsFile,BLs,BLsRest).
label_term(semantics(Pred),GramS,AbsFile,TokenS,['.'],BLs,[OpLine|BLsMid]) :-
  !,trunc_select(atom('semantics',OpLine,_),TokenS,TokenSMid),
  trunc_select(atom(Pred,_,_),TokenSMid,_),
  assertz(lsemantics(OpLine,Pred)),
  label_preds(GramS,AbsFile,BLs,BLsMid).
label_term(sub(_,_),GramS,AbsFile,_,['.'],BLs,BLsRest) :-  % skip these for now
  !,label_preds(GramS,AbsFile,BLs,BLsRest).
label_term(ext(_),GramS,AbsFile,_,['.'],BLs,BLsRest) :-
  !,label_preds(GramS,AbsFile,BLs,BLsRest).
label_term(cons(_,_),GramS,AbsFile,_,['.'],BLs,BLsRest) :-
  !,label_preds(GramS,AbsFile,BLs,BLsRest).

label_term(_,GramS,AbsFile,_,_,BLs,BLsRest) :-  % throw everything else out 
  !,label_preds(GramS,AbsFile,BLs,BLsRest).  %  (hook code etc. need to go in 
                                             %   a separate file)

label_rule_body((Op1,Op2),TokenS,(LabelledOp1,LabelledOp2),TokenSRest,
                BLs,BLsRest) :-
  !,label_rule_body(Op1,TokenS,LabelledOp1,TokenSMid,BLs,BLsMid),
  trunc_select(',',TokenSMid,TokenSMid2),
  label_rule_body(Op2,TokenSMid2,LabelledOp2,TokenSRest,BLsMid,BLsRest).
label_rule_body(cat> DtrDesc,TokenS,lcat(LabelledDtrDesc),TokenSRest,
                BLs,BLsRest) :-
  !,trunc_select(atom('cat',_,_),TokenS,TokenSMid),
  trunc_select(atom('>',_,_),TokenSMid,TokenSMid2),
  label_desc(DtrDesc,TokenSMid2,LabelledDtrDesc,_,TokenSRest,BLs,BLsRest).
label_rule_body(sem_head> DtrDesc,TokenS,lsem_head(LabelledDtrDesc),TokenSRest,
                BLs,BLsRest) :-
  !,trunc_select(atom('sem_head',_,_),TokenS,TokenSMid),
  trunc_select(atom('>',_,_),TokenSMid,TokenSMid2),
  label_desc(DtrDesc,TokenSMid2,LabelledDtrDesc,_,TokenSRest,BLs,BLsRest).
label_rule_body(cats> DtrDesc,TokenS,lcats(LabelledDtrDesc),TokenSRest,
                BLs,BLsRest) :-
  !,trunc_select(atom('cats',_,_),TokenS,TokenSMid),
  trunc_select(atom('>',_,_),TokenSMid,TokenSMid2),
  label_desc(DtrDesc,TokenSMid2,LabelledDtrDesc,_,TokenSRest,BLs,BLsRest).
label_rule_body(goal> GoalBody,TokenS,lgoal(LabelledGoalBody),TokenSRest,
                BLs,BLsRest) :-
  !,trunc_select(atom('goal',_,_),TokenS,TokenSMid),
  trunc_select(atom('>',_,_),TokenSMid,TokenSMid2),
  label_body(GoalBody,TokenSMid2,LabelledGoalBody,TokenSRest,BLs,BLsRest).
label_rule_body(sem_goal> GoalBody,TokenS,lsem_goal(LabelledGoalBody),
                TokenSRest,BLs,BLsRest):-
  !,trunc_select(atom('sem_goal',_,_),TokenS,TokenSMid),
  trunc_select(atom('>',_,_),TokenSMid,TokenSMid2),
  label_body(GoalBody,TokenSMid2,LabelledGoalBody,TokenSRest,BLs,BLsRest).

label_lr_body(morphs(NMBody,Morphs),TokenS,
              llrbody(LabelledNMBody,LabelledMorphs),InLine,BLs,BLsRest) :-
  label_lr_body_proper(NMBody,TokenS,LabelledNMBody,InLine,TokenSMid,
                       BLs,BLsMid),
  trunc_select(atom('morphs',_,_),TokenSMid,TokenSMid2),
  label_morphs(Morphs,TokenSMid2,LabelledMorphs,BLsMid,BLsRest).

label_lr_body_proper(if(NCBody,Cond),TokenS,
                     lif(LabelledNCBody,LabelledCond),InLine,TokenSRest,
                     BLs,BLsRest) :-
  !,label_lr_body_no_cond(NCBody,TokenS,LabelledNCBody,InLine,TokenSMid,
                          BLs,BLsMid),
  trunc_select(atom('if',_,_),TokenSMid,TokenSMid2),
  label_body(Cond,TokenSMid2,LabelledCond,TokenSRest,BLsMid,BLsRest).
label_lr_body_proper(NCBody,TokenS,LabelledNCBody,InLine,TokenSRest,
                     BLs,BLsRest):-
  label_lr_body_no_cond(NCBody,TokenS,LabelledNCBody,InLine,TokenSRest,
                        BLs,BLsRest).

label_lr_body_no_cond('**>'(DescIn,DescOut),TokenS,
                      llrmap(LabelledDescIn,LabelledDescOut),InLine,TokenSRest,
                      BLs,BLsRest) :-
  label_desc(DescIn,TokenS,LabelledDescIn,InLine,TokenSMid,BLs,BLsMid),
  trunc_select(atom('**>',_,_),TokenSMid,TokenSMid2),
  label_desc(DescOut,TokenSMid2,LabelledDescOut,_,TokenSRest,BLsMid,BLsRest).

label_morphs((Morph,Morphs),TokenS,lmorph_seq(LabelledMorph,LabelledMorphs),
             BLs,BLsRest) :-
  !,label_morph(Morph,TokenS,LabelledMorph,TokenSMid,BLs,BLsMid),
  trunc_select(',',TokenSMid,TokenSMid2),
  label_morphs(Morphs,TokenSMid2,LabelledMorphs,BLsMid,BLsRest).
label_morphs(Morph,TokenS,LabelledMorph,BLs,BLsRest) :-
  label_morph(Morph,TokenS,LabelledMorph,_,BLs,BLsRest).

label_morph(when(NCMorph,Cond),TokenS,
            lmorph_cond(WhenLine,LabelledNCMorph,Cond),TokenSRest,
            BLs,BLsRest) :-
  !,label_morph_no_cond(NCMorph,TokenS,LabelledNCMorph,TokenSMid,
                        BLs,[WhenLine|BLsRest]),
  trunc_select(atom('when',WhenLine,_),TokenSMid,TokenSMid2),
  source_debug:read(TokenSMid2,999,Cond,TokenSRest). % flush Cond (comma=1000)
label_morph(NCMorph,TokenS,LabelledNCMorph,TokenSRest,BLs,BLsRest) :-
  label_morph_no_cond(NCMorph,TokenS,LabelledNCMorph,TokenSRest,BLs,BLsRest).

label_morph_no_cond(becomes(StrIn,StrOut),TokenS,lmorph(OpLine,StrIn,StrOut),
                    TokenSRest,[OpLine|BLsRest],BLsRest) :-
  trunc_select(atom(becomes,OpLine,_),TokenS,TokenSMid), % this flushes StrIn
  source_debug:read(TokenSMid,949,StrOut,TokenSRest). % flush StrOut (when=950)

label_query(GoalDesc,Functor,Arity,LDList) :-
  functor(GoalDesc,Functor,Arity),
  GoalDesc =.. [_|ArgDescs],
  label_query_arg_list(ArgDescs,LDList).

label_query_arg_list([],[]).
label_query_arg_list([ArgDesc|ArgDescs],[LD|LDs]) :-
  label_query_arg(ArgDesc,LD),
  label_query_arg_list(ArgDescs,LDs).

label_query_arg(X,lvar(X,'query variable',none)) :-
  var(X),
  !.
label_query_arg(F:Desc,lfeat(F,none):LabelledDesc) :-
  !,label_query_arg(Desc,LabelledDesc).
label_query_arg([],lelist(none)) :-
  !.
label_query_arg([EltDesc|TailDesc],lnelist(none,none,LabelledED,LabelledTD)) :-
  !,label_query_arg(EltDesc,LabelledED),
  label_query_tail(TailDesc,LabelledTD).
label_query_arg((D1,D2),(LD1,LD2)) :-
  !,label_query_arg(D1,LD1),
  label_query_arg(D2,LD2).
label_query_arg((D1;D2),(LD1;LD2)) :-
  !,label_query_arg(D1,LD1),
  label_query_arg(D2,LD2).
label_query_arg((@ MacroCall),lmac(LabelledMC,none)) :-
  !,label_query_comp(MacroCall,lcomp(Functor,_,_,LabelledDescs)),
  LabelledMC =.. [Functor|LabelledDescs].
label_query_arg((Path1 == Path2),leq(Path1,Path2,none,LP1,LP2)) :-
  !,label_query_path(Path1,LP1),
  label_query_path(Path2,LP2).
label_query_arg((=\= Desc),lineq(none,LabelledDesc)) :-
  !,label_query_arg(Desc,LabelledDesc).
label_query_arg((a_ Atom),latom(Atom,none)) :-
  !.
label_query_arg(Type,ltype(Type,none)) :-
  atom(Type),
  !.
label_query_arg(Fun,lfun(Fun,LabelledFun)) :-
  label_query_comp(Fun,LabelledFun).

label_query_tail(X,lvar(X,'query variable',none)) :-
  var(X),
  !.
label_query_tail([],lelist(none)) :-
  !.
label_query_tail([EltDesc|TailDesc],
                 lnelist(none,none,LabelledED,LabelledTD)) :-
  !,label_query_arg(EltDesc,LabelledED),
  label_query_tail(TailDesc,LabelledTD).
label_query_tail(TailDesc,LabelledTailDesc) :-  % improper list
  label_query_arg(TailDesc,LabelledTailDesc).

label_query_comp(CompoundTerm,lcomp(Functor,Arity,none,LabelledDList)) :-
  CompoundTerm =.. [Functor|DescList],
  functor(CompoundTerm,_,Arity),
  label_query_arg_list(DescList,LabelledDList).

label_query_path([],[]).
label_query_path([Feat|Path],[lfeat(Feat,none)|LPath]) :-
  label_query_path(Path,LPath).

label_body((Body1,Body2),TokenS,(LabelledBody1,LabelledBody2),TokenSRest,
           BLs,BLsRest) :-
  !,label_body(Body1,TokenS,LabelledBody1,TokenSMid,BLs,BLsMid),
  trunc_select(',',TokenSMid,TokenSMid2),
  label_body(Body2,TokenSMid2,LabelledBody2,TokenSRest,BLsMid,BLsRest).
label_body((Body1;Body2),TokenS,(LabelledBody1;LabelledBody2),TokenSRest,
           BLs,BLsRest) :-
  !,label_body(Body1,TokenS,LabelledBody1,TokenSMid,BLs,BLsMid),
  trunc_select(';',TokenSMid,TokenSMid2),
  label_body(Body2,TokenSMid2,LabelledBody2,TokenSRest,BLsMid,BLsRest).
label_body(\+ Goal,TokenS,lneg(NegLine,LabelledGoal),TokenSRest,
           [NegLine|BLsMid],BLsRest):-
  !,trunc_select(atom('\+',NegLine,_),TokenS,TokenSMid),
  label_body(Goal,TokenSMid,LabelledGoal,TokenSRest,BLsMid,BLsRest).
label_body((Desc1 =@ Desc2),TokenS,lexteq(OpLine,LabelledD1,LabelledD2),
           TokenSRest,BLs,BLsRest) :-
  !,label_desc(Desc1,TokenS,LabelledD1,_,TokenSMid,BLs,[OpLine|BLsMid]),
  trunc_select(atom('=@',OpLine,_),TokenSMid,TokenSMid2),
  label_desc(Desc2,TokenSMid2,LabelledD2,_,TokenSRest,BLsMid,BLsRest).
label_body(prolog(PGoal),TokenS,lprolog(OpLine,PGoal),TokenSRest,
           [OpLine|BLsRest],BLsRest) :-
  !,trunc_select(atom('prolog',OpLine,_),TokenS,TokenSMid),
  trunc_select('(',TokenSMid,TokenSMid2),
  source_debug:read(TokenSMid2,1200,PGoal,TokenSMid3),  % flush PGoal
  trunc_select(')',TokenSMid3,TokenSRest).
label_body(!,TokenS,lcut,TokenSRest,BLs,BLs) :-
  !,trunc_select(atom('!',_,_),TokenS,TokenSRest).
label_body(true,TokenS,ltrue,TokenSRest,BLs,BLs) :-
  !,trunc_select(atom(true,_,_),TokenS,TokenSRest).
label_body(AtGoal,TokenS,LabelledAtGoal,TokenSRest,BLs,BLsRest) :-
  !,label_compound(AtGoal,TokenS,LabelledAtGoal,_,TokenSRest,BLs,BLsRest).

label_desc(X,TokenS,lvar(X,XName,XLine),XLine,TokenSRest,
           [XLine|BLsRest],BLsRest) :-
  var(X),
  !,var_trunc_select(X,TokenS,TokenSRest,XString,XLine),
  name(XName,XString).
label_desc((F:Desc),TokenS,(lfeat(F,FLine):LabelledDesc),FLine,TokenSRest,
           [FLine|BLsMid],BLsRest) :-
  !,trunc_select(atom(F,FLine,_),TokenS,TokenSMid),
  trunc_select(atom(':',_,_),TokenSMid,TokenSMid2),
  label_desc(Desc,TokenSMid2,LabelledDesc,_,TokenSRest,BLsMid,BLsRest).
label_desc([],TokenS,lelist(LBrackLine),LBrackLine,TokenSRest,
           [LBrackLine|BLsRest],BLsRest):-
  !,trunc_select(lbrack(LBrackLine),TokenS,TokenSMid),
  trunc_select(rbrack(_),TokenSMid,TokenSRest).
label_desc([EltDesc|TailDesc],TokenS,
           lnelist(LBrackLine,TailLine,LabelledEltDesc,LabelledTailDesc),
           LBrackLine,TokenSRest,[LBrackLine|BLsMid],BLsRest) :-
  !,trunc_select(lbrack(LBrackLine),TokenS,TokenSMid),
  label_desc(EltDesc,TokenSMid,LabelledEltDesc,_,TokenSMid2,BLsMid,BLsMid2),
  label_tail(TailDesc,TokenSMid2,LabelledTailDesc,TailLine,TokenSRest,
             BLsMid2,BLsRest).  % TailLine already added to BLs 
label_desc((Desc1,Desc2),TokenS,(LD1,LD2),FirstLine,TokenSRest,BLs,BLsRest) :-
  !,label_desc(Desc1,TokenS,LD1,FirstLine,TokenSMid,BLs,BLsMid),
  label_desc(Desc2,TokenSMid,LD2,_,TokenSRest,BLsMid,BLsRest).
label_desc((Desc1;Desc2),TokenS,(LD1;LD2),FirstLine,TokenSRest,BLs,BLsRest) :-
  !,label_desc(Desc1,TokenS,LD1,FirstLine,TokenSMid,BLs,BLsMid),
  label_desc(Desc2,TokenSMid,LD2,_,TokenSRest,BLsMid,BLsRest).
label_desc((@ MacroCall),TokenS,lmac(LabelledMC,OpLine),OpLine,
           TokenSRest,[OpLine|BLsMid],BLsRest) :-
  !,trunc_select(atom('@',OpLine,_),TokenS,TokenSMid),
  label_compound(MacroCall,TokenSMid,lcomp(Functor,_,_,LabelledDescs),_,
                 TokenSRest,BLsMid,BLsRest),
  LabelledMC =.. [Functor|LabelledDescs].
label_desc((Path1 == Path2),TokenS,leq(Path1,Path2,OpLine,LP1,LP2),
           LBrackLine,TokenSRest,BLs,BLsRest) :-
  !,trunc_select(lbrack(LBrackLine),TokenS,TokenSMid),
  label_path(Path1,TokenSMid,LP1,TokenSMid2,BLs,[OpLine|BLsMid]),
  trunc_select(atom('==',OpLine,_),TokenSMid2,TokenSMid3),
  trunc_select(lbrack(_),TokenSMid3,TokenSMid4),
  label_path(Path2,TokenSMid4,LP2,TokenSRest,BLsMid,BLsRest).
label_desc((=\= Desc),TokenS,lineq(OpLine,LabelledDesc),OpLine,TokenSRest,
           [OpLine|BLsMid],BLsRest) :-
  !,trunc_select(atom('=\=',OpLine,_),TokenS,TokenSMid),
  label_desc(Desc,TokenSMid,LabelledDesc,_,TokenSRest,BLsMid,BLsRest).
label_desc((a_ Atom),TokenS,latom(Atom,ALine),ALine,TokenSRest,
           [ALine|BLsRest],BLsRest) :-
  !,trunc_select(atom('a_',ALine,_),TokenS,TokenSMid),
  strip_immed_left_paren(TokenSMid,TokenSMid2),
  source_debug:read(TokenSMid2,999,Atom,TokenSRest). % flush Atom (comma=1000)
label_desc(Type,TokenS,ltype(Type,TLine),TLine,TokenSRest,
           [TLine|BLsRest],BLsRest) :-
  atom(Type),
  !,trunc_select(atom(Type,TLine,_),TokenS,TokenSRest).
label_desc(Fun,TokenS,lfun(Fun,LabelledFun),FirstLine,TokenSRest,BLs,BLsRest):-
  !,label_compound(Fun,TokenS,LabelledFun,FirstLine,TokenSRest,BLs,BLsRest).

label_path([],TokenS,[],TokenSRest,BLs,BLs) :-
  trunc_select(rbrack(_),TokenS,TokenSRest).
label_path([Feat|Path],TokenS,[lfeat(Feat,FLine)|LPath],TokenSRest,
           [FLine|BLsMid],BLsRest) :-
  trunc_select(atom(Feat,FLine,_),TokenS,TokenSMid),
  label_path(Path,TokenSMid,LPath,TokenSRest,BLsMid,BLsRest).

label_compound(CompoundTerm,TokenS,
               lcomp(Functor,Arity,FunLine,LabelledDescList),FunLine,
               TokenSRest,[FunLine|BLsMid],BLsRest) :-
  CompoundTerm =.. [Functor|DescList],
  functor(CompoundTerm,_,Arity),
  trunc_select(atom(Functor,FunLine,_),TokenS,TokenSMid),
  label_desc_list(DescList,TokenSMid,LabelledDescList,TokenSRest,
                  BLsMid,BLsRest).

label_tail(X,TokenS,lvar(X,XName,XLine),XLine,TokenSRest,
           [XLine|BLsRest],BLsRest):-
  var(X),
  !,var_trunc_select(X,TokenS,TokenSMid,XString,XLine),
  name(XName,XString),
  trunc_select(rbrack(_),TokenSMid,TokenSRest).
label_tail([],TokenS,lelist(RBrackLine),RBrackLine,TokenSRest,
           [RBrackLine|BLsRest],BLsRest):-
  !,trunc_select(rbrack(RBrackLine),TokenS,TokenSRest).
label_tail([EltDesc|TailDesc],TokenS,
           lnelist(HeadLine,TailLine,LabelledEltDesc,LabelledTailDesc),
           HeadLine,TokenSRest,BLs,BLsRest) :-
  !,trunc_select(',',TokenS,TokenSMid),
  label_desc(EltDesc,TokenSMid,LabelledEltDesc,HeadLine,TokenSMid2,BLs,BLsMid),
  label_tail(TailDesc,TokenSMid2,LabelledTailDesc,TailLine,TokenSRest,
             BLsMid,BLsRest).
label_tail(TailDesc,TokenS,LabelledTailDesc,HeadLine,TokenSRest,
           BLs,BLsRest) :-
  trunc_select('|',TokenS,TokenSMid),
  label_desc(TailDesc,TokenSMid,LabelledTailDesc,HeadLine,TokenSMid2,
             BLs,BLsRest),
  trunc_select(rbrack(_),TokenSMid2,TokenSRest).

label_desc_list([],TokenS,[],TokenS,BLs,BLs).
label_desc_list([Desc|DescList],TokenS,[LabelledDesc|LabelledDescList],
                TokenSRest,BLs,BLsRest) :-
  label_desc(Desc,TokenS,LabelledDesc,_,TokenSMid,BLs,BLsMid),
  label_desc_list(DescList,TokenSMid,LabelledDescList,TokenSRest,
                  BLsMid,BLsRest).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Utilities

d_chain_rule_body(lsem_head(_)).
d_chain_rule_body((lsem_head(_),_)) :-
  !.
d_chain_rule_body((_,Rest)) :-
  d_chain_rule_body(Rest).


d_chain_rule_body((lsem_goal(LGoalBefore),lsem_head(LSemHead),
                   lsem_goal(LGoalAfter),_Rest),
                  LGoalBefore,LSemHead,LGoalAfter) :-
  !.
d_chain_rule_body((lsem_goal(LGoalBefore),lsem_head(LSemHead),
                   lsem_goal(LGoalAfter)),
                  LGoalBefore,LSemHead,LGoalAfter) :-
  !.
d_chain_rule_body((lsem_head(LSemHead),lsem_goal(LGoalAfter),_Rest),
                  empty,LSemHead,LGoalAfter) :-
  !.
d_chain_rule_body((lsem_head(LSemHead),lsem_goal(LGoalAfter)),
                  empty,LSemHead,LGoalAfter) :-
  !.
d_chain_rule_body((lsem_goal(LGoalBefore),lsem_head(LSemHead),
                   _Rest),
                  LGoalBefore,LSemHead,empty) :-
  !.
d_chain_rule_body((lsem_goal(LGoalBefore),lsem_head(LSemHead)),
                  LGoalBefore,LSemHead,empty) :-
  !.
d_chain_rule_body((lsem_head(LSemHead),_Rest),
                  empty,LSemHead,empty) :-
  !.
d_chain_rule_body(lsem_head(LSemHead),
                  empty,LSemHead,empty) :-
  !.
d_chain_rule_body((_,Rest),LGoalBefore,LSemHead,LGoalAfter) :-
  d_chain_rule_body(Rest,LGoalBefore,LSemHead,LGoalAfter).


d_chain_rule_body((lsem_goal(LGoalBefore),lsem_head(LSemHead),
                   lsem_goal(LGoalAfter),Rest),
                  LGoalBefore,LSemHead,LGoalAfter,empty,Rest) :-
  !.
d_chain_rule_body((lsem_goal(LGoalBefore),lsem_head(LSemHead),
                   lsem_goal(LGoalAfter)),
                  LGoalBefore,LSemHead,LGoalAfter,empty,empty) :-
  !.
d_chain_rule_body((lsem_head(LSemHead),lsem_goal(LGoalAfter),Rest),
                  empty,LSemHead,LGoalAfter,empty,Rest) :-
  !.
d_chain_rule_body((lsem_head(LSemHead),lsem_goal(LGoalAfter)),
                  empty,LSemHead,LGoalAfter,empty,empty) :-
  !.
d_chain_rule_body((lsem_goal(LGoalBefore),lsem_head(LSemHead),
                   Rest),
                  LGoalBefore,LSemHead,empty,empty,Rest) :-
  !.
d_chain_rule_body((lsem_goal(LGoalBefore),lsem_head(LSemHead)),
                  LGoalBefore,LSemHead,empty,empty,empty) :-
  !.
d_chain_rule_body((lsem_head(LSemHead),Rest),
                  empty,LSemHead,empty,empty,Rest) :-
  !.
d_chain_rule_body(lsem_head(LSemHead),
                  empty,LSemHead,empty,empty,empty) :-
  !.
d_chain_rule_body((OtherDtr,Rest),LGoalBefore,LSemHead,LGoalAfter,
                  (OtherDtr,Others),NextDtrs) :-
  d_chain_rule_body(Rest,LGoalBefore,LSemHead,LGoalAfter,Others,NextDtrs).


capture_macro_vars(LD,MacroHead,LMD) :-
  MacroHead =.. [_|MacroVars],
  cmv_act(LD,MacroVars,LMD).

cmv_act(lvar(X,XName,XLine),MacroVars,NewX) :-
  !,(member_eq(X,MacroVars) -> 
      NewX = X
    ; NewX = lvar(X,XName,XLine)).
cmv_act(lnelist(HeadLine,TailLine,LH,LT),MacroVars,
        lnelist(HeadLine,TailLine,NewLH,NewLT)) :-
  !,cmv_act(LH,MacroVars,NewLH),
  cmv_act(LT,MacroVars,NewLT).
cmv_act(lineq(OpLine,LD),MacroVars,lineq(OpLine,NewLD)) :-
  !,cmv_act(LD,MacroVars,NewLD).
cmv_act(LF:LD,MacroVars,LF:NewLD) :-
  !,cmv_act(LD,MacroVars,NewLD).
cmv_act((LD1,LD2),MacroVars,(NewLD1,NewLD2)) :-
  !,cmv_act(LD1,MacroVars,NewLD1),
  cmv_act(LD2,MacroVars,NewLD2).
cmv_act((LD1;LD2),MacroVars,(NewLD1;NewLD2)) :-
  !,cmv_act(LD1,MacroVars,NewLD1),
  cmv_act(LD2,MacroVars,NewLD2).
cmv_act(lmac(MacroCall,OpLine),MacroVars,lmac(NewMacroCall,OpLine)) :-
  !,MacroCall =.. [Name|ArgDescs],
  cmv_act_list(ArgDescs,MacroVars,NewArgDescs),
  NewMacroCall =.. [Name|NewArgDescs].
cmv_act(lfun(_,lcomp(Functor,Arity,FunLine,LDList)),MacroVars,
        lfun(NewFun,lcomp(Functor,Arity,FunLine,NewLDList))) :-
  !,cmv_act_list(LDList,MacroVars,NewLDList),
  NewFun =.. [Functor|NewLDList].
cmv_act(Other,_,Other).

cmv_act_list([],_,[]).
cmv_act_list([LD|LDs],MacroVars,[NewLD|NewLDs]) :-
  cmv_act(LD,MacroVars,NewLD),
  cmv_act_list(LDs,MacroVars,NewLDs).

strip_immed_left_paren(['('|TokenS],TokenS) :-
  !. % if left paren doesn't immediately follow, delimiter will include space
strip_immed_left_paren(TokenS,TokenS).

% trunc_select/3 - like select/3 except only returns portion of remainder
%  that occurs after selected element.

trunc_select(X,[X|Rest],Rest) :-
  !.
trunc_select(X,[_Y|List],Rest) :-
  trunc_select(X,List,Rest).

var_trunc_select(X,[var(_,XLine,XString,Y)|TokenS],TokenS,XString,XLine) :-
  X == Y,
  !.
var_trunc_select(X,[_V|TokenS],TokenSRest,XString,XLine) :-
  var_trunc_select(X,TokenS,TokenSRest,XString,XLine).

my_read_tokens(GramS,TokenS,Vars) :-
  prolog:'$input'(Current,GramS),
  source_debug:xread_tokens(TokenS,Vars),
  prolog:'$input'(_,Current).

skip_count_translate([]).
skip_count_translate([N|BLs]) :-
  sct_act(BLs,N,0).

sct_act([],N,Expected) :-
  Skip is N - Expected,
  write(user_error,Skip),space,write(user_error,1),space.
sct_act([NextN|BLs],N,Expected) :-
  Skip is N - Expected,
  NewExpect is N + 1,
  write(user_error,Skip),space,
  sc_count(BLs,NextN,NewExpect,1).

sc_count([],N,N,Accum) :-
  !,LastCount is Accum + 1,
  write(user_error,LastCount),space.
sc_count([],N,Expected,Accum) :-
  Skip is N - Expected + 1,
  write(user_error,Accum),space,write(user_error,Skip),space,
  write(user_error,1),space.
sc_count([NextN|BLs],N,N,Accum) :-
  !,NewAccum is Accum + 1,
  NewExpect is N + 1,
  sc_count(BLs,NextN,NewExpect,NewAccum).
sc_count([NextN|BLs],N,Expected,Accum) :-
  Skip is N - Expected + 1,
  NewExpect is N + 1,
  write(user_error,Accum),space,write(user_error,Skip),space,
  sc_count(BLs,NextN,NewExpect,1).

abolish_lpreds :-
  abolish('l--->',3),abolish(lif,2),abolish('l+++>',2),abolish(llr,3),
  abolish(lrule,4),abolish(lmacro,2),abolish(lempty,2),
  abolish(ale_debugging,1),abolish(ale_break,1),abolish(ale_breakable_lines,1).

space :- write(user_error,' ').

tread(T,V) :-
  hstream(S),
  prolog:'$input'(Curr,S),
  source_debug:xread_tokens(T,V),
  assert(foo(T,V)),
  prolog:'$input'(_,Curr).

% V contains a mapping of RefVars to char codes for each instance of a 
% *named* var

tparse(T,Term,Left) :-
  source_debug:read(T,1200,Term,Left).

tinit(F,S) :-
  open(F,read,S),
  assert(hstream(S)).

:- nl,write('
ALE Source-Level Debugger Version 2.0Beta; July, 1998
Copyright (C) 1998, Gerald Penn
All rights reserved'),nl.
